#uses "fwGeneral/fwDPELock.ctl"
#uses "classes/unModbusRedirect/UnModbusRedirectConfiguration.ctl"


class UnModbusRedirectClient {

	// names of #event in the panel that will be triggered
	// they need to be uiConnect'ed
    public const string hasLockEvent="hasLockChanged";
    public const string hasServerEvent="hasServerChanged";
    public const string cmdSubmittedEvent="cmdSubmitted";
    public const string cmdErrorEvent="cmdError";
    public const string cmdResponseEvent="cmdResponse";
	public const string cfgFileChangedEvent="cfgFileChanged";
	public const string cfgCmdChangedEvent="cfgCmdChanged";


    public const string dp="unModbusRedirect";

    private static shared_ptr<UnModbusRedirectClient> m_instance=nullptr;
    private UnModbusRedirectClient() { /* restricted constructor - use getInstance() factory pattern */}

    public bool lockOK   = false;
    public bool serverOK = false;
    public string clientLocker="";
    public string serverLocker="";

    public static shared_ptr<UnModbusRedirectClient> getInstance()
    {
      if (equalPtr(m_instance,nullptr)) {
        m_instance = new UnModbusRedirectClient();
        m_instance.initialize();
      }
      return m_instance;
    }

    private void initialize() { /* the functionality moved to connect() that must be called by the client */ }

	public void connect()
	{
      dpConnect(m_instance, m_instance.ClientLockCB,    TRUE,_fwDPELock_getLockConfig(m_instance.dp+".command" )+"._locked");
      dpConnect(m_instance, m_instance.ServerLockCB,    TRUE,_fwDPELock_getLockConfig(m_instance.dp+".response")+"._locked");

      dpConnect(m_instance, m_instance.CfgFileChangedCB,TRUE,m_instance.dp+".configFile",m_instance.dp+".configFile:_offline.._invalid");
      dpConnect(m_instance, m_instance.CmdChangedCB,    TRUE,m_instance.dp+".redirectModbusCmd",m_instance.dp+".redirectModbusCmd:_offline.._invalid");

      delay(0,100); // make time for initial callbacks to fire up
	}


	protected void CfgFileChangedCB(string dpe, string newCfgFile, string dpeInvalid, bool invalid)
	{
		delay(0,100); // let the server react first
		triggerEvent(cfgFileChangedEvent, newCfgFile, !invalid);
	}

	protected void CmdChangedCB(string dpe, string newCmd, string dpeInvalid, bool invalid)
	{
		delay(0,100); // let the server react first
		triggerEvent(cfgCmdChangedEvent, newCmd, !invalid);	
	}

	/**
		Called due to either the server obtaining or loosing the lock,
		which usually means that it starts or stops.
		
		However, the sole fact of having the lock is not sufficient to determine
		that it is ready for operation! It may still be initializing
		(e.g. waiting for its own dpConnects to stabilize).
		It will set the async response "STARTING" (just after obtaining the lock),
		and then "STARTED" when it is ready for operation.
		We should be waiting for these values.
		
	*/
    protected synchronized void ServerLockCB(string dpeLock, bool locked)
    {		
		// reset the state
		serverLocker="";
		serverOK=false;

  		if (!locked) { // either not yet started (and this is the first callback) or lost the server
			triggerEvent(hasServerEvent, false);
			return;
  		}
    
    	// process must be running, as the DPE is locked
    	// we need to see if it was already running
    	// (and this is clients' first start) and ready for operation
    	// or it has just started and proceeds with initiaization
    	// in which case we need to wait till it is done
    	string responseDPE       = m_instance.dp+".response";
    	string responseLockedDPE = _fwDPELock_getLockConfig(responseDPE)+"._locked";
    	responseDPE             += ":_original.._value";
    	dyn_anytype retValues;
    	time timeout=1; // wait for 1 second
    	bool timeExpired;
    	
    	dpWaitForValue( makeDynString(responseDPE),// we wait for value on this one
    					makeDynAnytype("STARTING"),               		// the value we expect
    					makeDynString(responseDPE,responseLockedDPE),	// values we want to get returned at this point
    					retValues, timeout, timeExpired);
    	
    	// if timer expired the server was already running
    	// otherwise wait for its startup...    	
    	if (!timeExpired) {
    		// means this is the startup sequence and we need to wait for finalization of startup
    		timeout=1;
    		dpWaitForValue( makeDynString(responseDPE),// we wait for value on this one
    						makeDynAnytype("STARTED"),               		// the value we expect
    						makeDynString(responseDPE,responseLockedDPE),	// values we want to get returned at this point
    						retValues, timeout, timeExpired);
    		if (timeExpired) {
    			// we did not have a startup confirmation within a time limit
    			DebugTN(__FUNCTION__,"WE DID NOT GET CONFIRMATION OF SERVER STARTUP");
				triggerEvent(hasServerEvent, false);
				return;
    		}
    	}
    	
    	// at this point we surely have the server
	    dyn_string lockDetails, exceptionInfo;
	    fwDPELock_getLocked(dpSubStr(dpeLock,DPSUB_SYS_DP_EL), lockDetails, exceptionInfo);
	    if (dynlen(exceptionInfo)) {
			DebugTN(__FUNCTION__,exceptionInfo);
			triggerEvent(hasServerEvent, false);
	    }
	    
		serverLocker = lockDetails[fwDPELock_LOCK_MANAGER_DETAIL]  + " #"+
                       lockDetails[fwDPELock_LOCK_MANAGER_REPLICA] + " ("+
                       lockDetails[fwDPELock_LOCK_MANAGER_HOST]    + ")";
		serverOK=true;
		triggerEvent(hasServerEvent, true);
    }

    /**
	Logic:
	 - if locked and we own the lock: trigger with hasLock=TRUE
	 - if locked but not our lock: trigger with hasLock=FALSE and clientLocker not-empty
	   (indicates to whom the lock belongs)
	 - if unlocked: trigger with hasLock=FALSE and clientLocker set to empty
    */
    protected void ClientLockCB(string dpe, bool locked)
    {
      dyn_string lockDetails, exceptionInfo;
      bool hasLock=locked;
      if (!locked) {
        clientLocker="";
	    triggerEvent( hasLockEvent,false);
        return;
      }
      clientLocker="(unknown)";
      fwDPELock_getLocked(dpSubStr(dpe,DPSUB_SYS_DP_EL), lockDetails, exceptionInfo);
      if (!dynlen(exceptionInfo)) {
        clientLocker = lockDetails[fwDPELock_LOCK_USER_NAME]       + "@" +
                       lockDetails[fwDPELock_LOCK_MANAGER_DETAIL]  + " #"+
                       lockDetails[fwDPELock_LOCK_MANAGER_REPLICA] + " ("+
                       lockDetails[fwDPELock_LOCK_MANAGER_HOST]    + ")";
      }

      bool hasTheLock = (myManId() == lockDetails[fwDPELock_LOCK_MANAGER_MANID]);
      triggerEvent( hasLockEvent, hasTheLock);
    }


    public bool tryLock(int lockOwnershipTimeout=60)
    {
      dyn_string exceptionInfo;
      bool gotTheLock=fwDPELock_tryLock(dp+".command", exceptionInfo, 3, lockOwnershipTimeout);
      if (dynlen(exceptionInfo)) {
        DebugTN(__FILE__,__FUNCTION__,exceptionInfo);
        return false;
      }
      return true;
    }

    public void releaseLock()
    {
      dyn_string lockDetails, exceptionInfo;
      bool locked = fwDPELock_getLocked(dp+".command", lockDetails, exceptionInfo);
      if (dynlen(exceptionInfo)) {
        DebugTN(__FILE__,__FUNCTION__,exceptionInfo);
      }

      if (!locked) return;

      if (myManId() == lockDetails[fwDPELock_LOCK_MANAGER_MANID]) {
        fwDPELock_unlock(dp+".command",exceptionInfo);
        if (dynlen(exceptionInfo)) {
          DebugTN(__FILE__,__FUNCTION__,exceptionInfo);
        }
      }
    }

    public synchronized string sendCommand(string cmd, dyn_string params=makeDynString())
    {
	if (serverLocker=="") {
	    triggerEvent(cmdErrorEvent, "NO SERVER");
	    return "ERROR#SendCommand:NO_SERVER";
	}
    
	if (!tryLock()) {
	    DebugTN(__FILE__,"Cannot send the command - tryLock() failed", cmd);
	    triggerEvent(cmdErrorEvent, "NO LOCK");
	    return "ERROR#SendCommand:NO_LOCK";
	}
	triggerEvent(cmdSubmittedEvent,cmd);
	string cmdString=cmd;
	if (dynlen(params)) {
	    cmdString+="#"+params[1];
	    for (int i=2;i<=dynlen(params);i++) cmdString+="|"+params[i];
	}
	dyn_anytype retVals;
	bool timeoutExpired;
	int timeout=5;
	dpSetAndWaitForValue(makeDynString(dp+".command:_original.._value"),makeDynAnytype(cmdString), // values to set
			     makeDynString(dp+".response:_original.._value"),makeDynAnytype(), // conditions to wait
			     makeDynString(dp+".response:_original.._value"), retVals,           // return values
			     timeout, timeoutExpired);
	if (timeoutExpired) {
	    triggerEvent(cmdErrorEvent, "NO RESPONSE");
	    return "ERROR#SendCommand:NO_RESPONSE";
	}
	string retString;
	if (dynlen(retVals)) retString=retVals[1];
	if (strpos(retString,"ERROR#")==0) triggerEvent(cmdErrorEvent, substr(retString,strlen("ERROR#")-1));
	triggerEvent(cmdResponseEvent,retString);
	return retString;
    }
    
    public void sendConfiguration(UnModbusRedirectConfiguration &cfg)
    {
    	string s=cfg.serialize();
    	sendCommand("SETCONFIGURATION",s);
    }
};
