/** @file unCMWServerStatus.ctl */
#uses "unMessageText.ctl"
#uses "fwGeneral/fwException.ctl"
#uses "unCMWServer.ctl"

/** @name Class definition */
//@(

/**
 *  @class CMWServerStatus
 *  <b>Thic ic CTRL++, not C++  !</b><br>
 *  The purpose of this class is calculation of alarm conditions for system intergity.
 *  Each instance conatins all necessary data for one CMWServer manager, plus methods
 *  for calculating the final result.<br>
 *  In addition, each instance is able to display all details of CMWServer manager state
 *  in row of Table control on panel, the names of columns in the table are <b>hardcoded</b>.<br>
 *  This class also provides static methods for adding/removing CMW server managers to monitoring.<br>
 *  The main idea for developing this class was to collect in one place all functionality, previously
 *  spread over deveral global variable and functions in different CTRL libraries.
 */

/**
 * Final alarm state. Only few values are defined here, other values are built as combination
 * (bitwise OR) of single bits corresponding to individual alarms in CMWServerBoolState
 */
enum CMWServerAlarm
{
  OK = 0,                   ///< No alarm
  Stopped = 8193,           ///< Manager is not running (no mamanger number in _Connections.Api.ManNums)
  Blocked = 8194,           ///< Manager is 'blocked' (no heartbeat)
  ServerNotRunning = 8195   ///< Manager is running, but CMW server was not created/started
};

/**
 * Enum used to identify individual state DPEs of type bool - one dpConnect() callback is used for all,
 * the value of this enum is used as userData in callback.
 * The 'final' alarm value (integer) is calcuiated as combination (bitwise OR) of single bits, corresponding
 * to every individual alarm. For example, if ValueQueueOverflow and RbacDisableIndicator alarms are active,
 * then resulting value will be: (1 << ValueQueueOverflow) | (1 << RbacDisableIndicator).
 * The very first alarm defines the very first bit number used for such calculation.
 * Note that only resulting values above 10 are considered to be alarms, this is the reason for starting
 * enum values from 4, not from 0.
 */
enum CMWServerBoolState
{
  ValueQueueOverflow = 4,     ///< queue for new DPE values has overflown: Alarms.ValueQueueOverflow
  SetRequestQueueOverflow,    ///< queue for SET requests has overflown: Alarms.SetRequestQueueOverflow
  SetResponseQueueOverflow,   ///< queue for SET responses has overflown: Alarms.SetResponseQueueOverflow
  RbacDisableIndicator,       ///< RBAC is disabled: Alarms.RbacDisabled
  DpeConnectionLost,          ///< Connection to one or more DPEs was lost while server is running: Alarms.DpeConnectionLost
  ConfigFileInvalid,          ///< Config file is not valid while server is running: Alarms.ConfigFileInvalid
  Running = 8192              ///< CMW server was started by manager: .Status.Running
};

/// To switch ON debug messages
private const string C_DEBUG = "CMWServerStatus";

class CMWServerStatus
{
  /************************************ Statis methods *******************************************************************/

  /**
   * @public
   *
   * Add manager to the monitoring, used by both CTRL and UI managers. Add necessary instance members,
   * make relevant dpConnect() calls... i.e. do everything to be able to report te 'final' state for
   * system integrity (alarms) for given CMWServer manager.
   *
   *  @param sSystem System name
   *  @param sManagerDp The name of CMWServer manager configuration DP
   *  @param sAlarmDp The name of DP where resulting alarm state for manager shall be written
   *  @param exceptionInfo Any detected errors will be added to the content of this variable
   *
   *  @return true in case of success
   */
  public static bool startManagerMonitoring(string sSystem, string sManagerDp, string sAlarmDp, dyn_string &exceptionInfo)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + sManagerDp + "), total " + mappinglen(g_mInstances));
    shared_ptr<CMWServerStatus> instance;
    bool bVeryFirst;
    synchronized(g_mInstances)
    {
      if(mappingHasKey(g_mInstances, sManagerDp))
      {
        DebugFTN(C_DEBUG, __FUNCTION__ + "(" + sManagerDp + "), already monitored");
        return true;  // Manager is already monitored
      }
      DebugFTN(C_DEBUG, __FUNCTION__ + "(" + sManagerDp + "), adding new");
      bVeryFirst = mappinglen(g_mInstances) == 0;
      instance = new CMWServerStatus(sSystem, sManagerDp, sAlarmDp, exceptionInfo);
      g_mInstances[sManagerDp] = instance;
    }
    instance.connect(exceptionInfo);
    bool bResult = instance.m_bInitialized;
    if(bVeryFirst)  // If this is the very first manager - connect to 'global' DPE and start heartbeat check thread
    {
      DebugFTN(C_DEBUG, __FUNCTION__ + "(" + sManagerDp + "), the very first, adding global controls");
      dpConnect("unCMWServerStatus_AllApiNums_CB", sSystem + "_Connections.Api.ManNums");
      dyn_errClass err = getLastError();
      if(dynlen(err) > 0)
      {
        fwException_raise(exceptionInfo, "ERROR", __FUNCTION__ + "(): dpConnect failed for _Connections.Api.ManNums", "");
      }
      g_iThreadId = startThread("unCMWServerStatus_HeartBeatCheck");
    }
    else
    {
      DebugFTN(C_DEBUG, __FUNCTION__ + "(" + sManagerDp + "), not the very first");
      instance.processManConnection(g_diAllApiManNums);  // Apply namager numbers list for new instance
    }
    instance.m_tConfigFileCheckTime = getCurrentTime();
    return bResult;
  }

  /**
   * @public
   *
   * Stop monitoring for single manager, used by CTRL and UI. Note that system name is not
   * passed in arguments = each instance remembers to what system it was connected.
   *
   *  @param sManagerDp The name of CMWServer manager configuration DP
   *  @param exceptionInfo Any detected errors will be added to the content of this variable
   */
  public static void stopManagerMonitoring(string sManagerDp, dyn_string &exceptionInfo)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + sManagerDp + "), total " + mappinglen(g_mInstances));
    shared_ptr<CMWServerStatus> instance;
    bool bVeryLast;
    synchronized(g_mInstances)
    {
      if(!mappingHasKey(g_mInstances, sManagerDp))
      {
        DebugFTN(C_DEBUG, __FUNCTION__ + "(" + sManagerDp + "), not found in mapping");
        return;  // Manager is not monitored
      }
      DebugFTN(C_DEBUG, __FUNCTION__ + "(" + sManagerDp + "), removing");
      instance = g_mInstances[sManagerDp];
      mappingRemove(g_mInstances, sManagerDp);
      bVeryLast = mappinglen(g_mInstances) == 0;
    }
    instance.disconnect(exceptionInfo);
    if(bVeryLast)  // If this is the very last manager - disconnect from 'global' DPE and stop heartbeat check thread
    {
      DebugFTN(C_DEBUG, __FUNCTION__ + "(" + sManagerDp + "), the very last, disconnecting global");
      dpDisconnect("unCMWServerStatus_AllApiNums_CB", instance.m_sSystem + "_Connections.Api.ManNums");
      dyn_errClass err = getLastError();
      if(dynlen(err) > 0)
      {
        fwException_raise(exceptionInfo, "ERROR", __FUNCTION__ + "(): dpDisconnect failed for _Connections.Api.ManNums", "");
      }
      if(g_iThreadId >= 0)
      {
        stopThread(g_iThreadId);
        g_iThreadId = -1;
      }
    }
  }

  /**
   * @public
   *
   * Get alarm DP names for all currently monitored managers. Note: not the names of CMWServer
   * manager confirurations DPs. but the names of DPs where resulting alarms are reported.
   *
   *  @return The list of alarm DP names.
   */
  public static dyn_string getAllAlarms()
  {
    dyn_string dsResult;
    synchronized(g_mInstances)
    {
      int iTotal =  mappinglen(g_mInstances);
      for(int idx = 1 ; idx <= iTotal ; idx++)
      {
        shared_ptr<CMWServerStatus> instance = mappingGetValue(g_mInstances, idx);
        if(instance.m_sAlarmDpName != "")
        {
          dynAppend(dsResult, instance.m_sAlarmDpName);
        }
      }
    }
    return dsResult;
  }

  /**
   * @public
   *
   * Completely clean monitoring pool, used by UI when opening panel, possibly after previously opened panel was closed
   * Note that here we only need to clean up the memory, results of dpConnect() and startThread() were 'lost' when
   * previously opened panel was closed. The method shall be used in UI only.
   */
  public static void clean()
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(), total " + mappinglen(g_mInstances) + ", they are:", g_mInstances);
    mappingClear(g_mInstances);
    dynClear(g_diAllApiManNums);
    g_iThreadId = -1;
    DebugFTN(C_DEBUG, __FUNCTION__ + "() FINISH");
  }

  /**
   * @public
   *
   * Process data arrived in callback for the list of all connected API managers in the system, from _Connections.Api.ManNums
   *
   * @param diAllManNums The numbers of all connected API managers
   */
  public static void allApiManNumsCallback(const dyn_int &diAllManNums)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(), new list is:", diAllManNums);
    g_diAllApiManNums = diAllManNums;
    synchronized(g_mInstances)
    {
      for(int idx = mappinglen(g_mInstances) ; idx > 0 ; idx--)
      {
        shared_ptr<CMWServerStatus> instance = mappingGetValue(g_mInstances, idx);
        instance.processManConnection(g_diAllApiManNums);
      }
    }
    DebugFTN(C_DEBUG, __FUNCTION__ + "() FINISH");
  }

  /**
   * @public
   *
   * Process data arrived in callback value for heartbeat period
   *
   *  @param sManagerDp The name of CMWServer configuration DP == source of callback
   *  @param iPeriod New heartbeat period value
   */
  public static void heartBeatPeriodCallback(string sManagerDp, int iPeriod)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + sManagerDp + "," + iPeriod + ")");
    shared_ptr<CMWServerStatus> instance = findInstance(sManagerDp);
    if(instance != nullptr)
    {
      instance.m_iHeartBeatPeriod = iPeriod == 0 ? 1000 : iPeriod;  // 1000 msec = default period
    }
    DebugFTN(C_DEBUG, __FUNCTION__ + "() FINISH");
  }

  /**
   * @public
   *
   * Process heartbeat callback, next hertbeat time has arrived
   *
   *  @param sManagerDp The name of CMWServer configuration DP == source of callback
   *  @param tHeartBeat New heartbeat time for this manager
   */
  public static void heartBeatTimeCallback(string sManagerDp, time tHeartBeat)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + sManagerDp + "), time is", tHeartBeat);
    shared_ptr<CMWServerStatus> instance = findInstance(sManagerDp);
    if(instance != nullptr)
    {
      instance.processHeartBeatTime(tHeartBeat);
    }
    DebugFTN(C_DEBUG, __FUNCTION__ + "() FINISH");
  }

  /**
   * @public
   *
   * Process new value arrived from callback for one of bool DPEs has arrived.
   * Note that connection was done using dpConnectUserData, so here we have extra argument
   * telling which of boolean DPEs has changed
   *
   *  @param sManagerDp The name of CMWServer configuration DP == source of callback
   *  @param iValueType Enumerion value allowing to identify which if DPE values has arrived here
   *  @param bState New value of DPE
   */
  public static void boolStateCallback(string sManagerDp, CMWServerBoolState iValueType, bool bState)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + sManagerDp + ", " + iValueType + ", " + bState + ")");
    shared_ptr<CMWServerStatus> instance = findInstance(sManagerDp);
    if(instance != nullptr)
    {
      switch(iValueType)
      {
      case CMWServerBoolState::Running:
        instance.processSrvRunning(bState);
        break;
      case CMWServerBoolState::ValueQueueOverflow:
        instance.processValueQueueOverflow(bState);
        break;
      case CMWServerBoolState::SetRequestQueueOverflow:
        instance.processSetRequestQueueOverflow(bState);
        break;
      case CMWServerBoolState::SetResponseQueueOverflow:
        instance.processSetResponseQueueOverflow(bState);
        break;
      case CMWServerBoolState::RbacDisableIndicator:
        instance.processRbacDisabled(bState);
        break;
      case CMWServerBoolState::DpeConnectionLost:
        instance.processDpeConnectionLost(bState);
        break;
      case CMWServerBoolState::ConfigFileInvalid:
        instance.processConfigFileInvalid(bState);
        break;
      }
    }
    DebugFTN(C_DEBUG, __FUNCTION__ + "() FINISH");
  }

  /**
   * @public
   *
   * Find instance for given manager DP name in the list of registered instances, used
   * mainly to find out which instance shall be informed about new DPE callback value.
   *
   *  @param sManagerDp The name of CMWServer configuration DP for which instance shall be found
   *
   *  @return Pointer to the instance found, may be @c nullptr if not found
   */
  private static shared_ptr<CMWServerStatus> findInstance(string sManagerDp)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + sManagerDp + ")");
    shared_ptr<CMWServerStatus> result;
    synchronized(g_mInstances)
    {
      if(mappingHasKey(g_mInstances, sManagerDp))
      {
        result = g_mInstances[sManagerDp];
      }
    }
    DebugFTN(C_DEBUG, __FUNCTION__ + "() FINISH, result is: " + (result == nullptr ? "NOT FOUND" : "found"));
    return result;
  }

  /**
   * @public
   *
   * Check hertbeat 'state' for all managers, i.e. check for every manager if heartbeat has arrived
   * at expected interval; otherwise the manager is considered to be dead.<br>
   * This function also calls perodically the check of configuration file for running managers.
   */
  public static void checkAllHeartBeats()
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "()");
    shared_ptr<CMWServerStatus> instanceToCheckFile;
    time tLastFileCheck = getCurrentTime();
    synchronized(g_mInstances)
    {
      for(int idx = mappinglen(g_mInstances) ; idx > 0 ; idx--)
      {
        shared_ptr<CMWServerStatus> instance = mappingGetValue(g_mInstances, idx);
        instance.calculateApplyState();
        if(instance.m_bCmwRunning)
        {
          if(instance.m_tConfigFileCheckTime < tLastFileCheck)
          {
            tLastFileCheck = instance.m_tConfigFileCheckTime;
            instanceToCheckFile = instance;
          }
        }
      }
    }
    // Check config file for running manager, only when running in CTRL manager
    if((instanceToCheckFile != nullptr) && (myManType() == CTRL_MAN))
    {
      time now = getCurrentTime();
      if((now - tLastFileCheck) > 10)  /// Do not check too often
      {
        instanceToCheckFile.checkConfigFile(now);
      }
    }
    DebugFTN(C_DEBUG, __FUNCTION__ + "() FINISH");
  }

  /************************************ Statis members *******************************************************************/
  
  /// Mapping containing all instances. Key = manager DP name, value = shared_ptr to instance
  private static mapping g_mInstances;

  /// List of all connected API managers in the system, from _Connections.Api.ManNums
  private static dyn_int g_diAllApiManNums;

  /// ID of thread performing heartbeat monitoring for all managers
  private static int g_iThreadId = -1;

  /********************* Instance methods, one instance == one CMWServer manager in the system ***************************/
  /**
   * The only constructor, initialize instance members
   *
   *  @param sSystem System name
   *  @param sManagerDp The name of CMWServer manager configuration DP
   *  @param sAlarmDp The name of DP where resulting alarm state for manager shall be written
   *  @param exceptionInfo Any detected errors will be added to the content of this variable
   */
  private CMWServerStatus(string sSystem, string sManagerDp, string sAlarmDp, dyn_string &exceptionInfo)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + sManagerDp + ", " + sAlarmDp + ")");
    m_sManagerDp = sManagerDp;
    m_sSystem = sSystem;
    m_iManNum = unCMWServer_decodeManNum(sManagerDp, exceptionInfo);
    m_tAlarmTime = getCurrentTime();
    m_tHeartBeatTime = makeTime(2000, 1, 1);  // Far in the past
    m_sAlarmDpName = sAlarmDp;
    DebugFTN(C_DEBUG, __FUNCTION__ + "() FINISH");
  }
  
  /**
   * Connect to all essential DPEs of manager DP. Afer this call instance became 'alive',
   * i.e. the internal state of instance really reflects the state of corresponding CMWServer manager.
   *
   *  @param exceptionInfo Any detected errors will be added to the content of this variable
   */
  private void connect(dyn_string &exceptionInfo)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ")");
    string sManagerDp = m_sSystem + m_sManagerDp;
    // Heartbeat period
    dpConnect("unCMWServerStatus_HeartBeatPeriod_CB", sManagerDp + ".Settings.man_NumPeriod");
    dyn_errClass err = getLastError();
    if(dynlen(err) > 0)
    {
      fwException_raise(exceptionInfo, "ERROR", __FUNCTION__ + "(): dpConnect() failed for " +
                        sManagerDp + ".Settings.man_NumPeriod", "");
      DebugFTN(C_DEBUG, __FUNCTION__ + "() dpConnect() failed for " + sManagerDp + ".Settings.man_NumPeriod");
      return;
    }

    // Heartbeat time
    dpConnect("unCMWServerStatus_HeartBeatTime_CB", sManagerDp + ".Settings.man_Num:_online.._stime");
    err = getLastError();
    if(dynlen(err) > 0)
    {
      fwException_raise(exceptionInfo, "ERROR", __FUNCTION__ + "(): dpConnect() failed for " +
                        sManagerDp + ".Settings.man_Num:_online.._stime", "");
      DebugFTN(C_DEBUG, __FUNCTION__ + "() dpConnect() failed for " + sManagerDp + ".Settings.man_Num:_online.._stime");
      return;
    }
  
    // Running state of CMW Server plus state of 3 alarms: all DPEs are of type bool, so use single callback for all
    // in order to distinguish - use userData = enumn corresponding to instance members. If one dpConnect failed - do
    // not continue with connections
    if(connectToBoolState(sManagerDp + ".Status.Running", CMWServerBoolState::Running, exceptionInfo)
      && connectToBoolState(sManagerDp + ".Alarms.ValueQueueOverflow", CMWServerBoolState::ValueQueueOverflow, exceptionInfo)
      && connectToBoolState(sManagerDp + ".Alarms.SetRequestQueueOverflow", CMWServerBoolState::SetRequestQueueOverflow, exceptionInfo)
      && connectToBoolState(sManagerDp + ".Alarms.SetResponseQueueOverflow", CMWServerBoolState::SetResponseQueueOverflow, exceptionInfo)
      && connectToBoolState(sManagerDp + ".Alarms.RbacDisabled", CMWServerBoolState::RbacDisableIndicator, exceptionInfo)
      && connectToBoolState(sManagerDp + ".Alarms.DpeConnectionLost", CMWServerBoolState::DpeConnectionLost, exceptionInfo)
      && connectToBoolState(sManagerDp + ".Alarms.ConfigFileInvalid", CMWServerBoolState::ConfigFileInvalid, exceptionInfo))
    {
      m_bInitialized = m_bForceDisplay = true;
    }
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + sManagerDp + ") FINISH, initialized " + m_bInitialized);
  }

  /**
   * Connect to one status DPE of type bool. All boolean DPEs use single callback callback function,
   * the argument for dpConnectUserData() is used to identify what DPE was the source of callback.
   *
   *  @param sDpeName The name of DPE to connect to
   *  @param eValueType The userData argument for dpConnectUserData(), will be used in callback processing
   *  @param exceptionInfo Any detected errors will be added to the content of this variable
   *
   *  @return true in case of success   
   */
  private bool connectToBoolState(string sDpeName, CMWServerBoolState eValueType, dyn_string &exceptionInfo)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + sDpeName + ", " + eValueType + ")");
    dpConnectUserData("unCMWServerStatus_BoolState_CB", (int)eValueType, sDpeName);
    dyn_errClass err = getLastError();
    if(dynlen(err) > 0)
    {
      fwException_raise(exceptionInfo, "ERROR", __FUNCTION__ + "(): dpConnectUserData() failed for " + sDpeName, "");
      DebugFTN(C_DEBUG, __FUNCTION__ + "(" + sDpeName + ", " + eValueType + ") dpConnectUserData() failed");
      return false;
    }
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + sDpeName + ", " + eValueType + ") FINISH OK");
    return true;
  }

  /**
   * Disconnect from all essential DPEs of manager state, normally this is called when given CMWServer
   * manager's state doesn't need to be monitored
   *
   *  @param exceptionInfo Any detected errors will be added to the content of this variable
   */
  private void disconnect(dyn_string &exceptionInfo)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ")");
    string sManagerDp = m_sSystem + m_sManagerDp;
    m_bInitialized = false;
    disconnectFromBoolState(sManagerDp + ".Alarms.ConfigFileInvalid", CMWServerBoolState::ConfigFileInvalid, exceptionInfo);
    disconnectFromBoolState(sManagerDp + ".Alarms.DpeConnectionLost", CMWServerBoolState::DpeConnectionLost, exceptionInfo);
    disconnectFromBoolState(sManagerDp + ".Alarms.RbacDisabled", CMWServerBoolState::RbacDisableIndicator, exceptionInfo);
    disconnectFromBoolState(sManagerDp + ".Alarms.SetResponseQueueOverflow", CMWServerBoolState::SetResponseQueueOverflow, exceptionInfo);
    disconnectFromBoolState(sManagerDp + ".Alarms.SetRequestQueueOverflow", CMWServerBoolState::SetRequestQueueOverflow, exceptionInfo);
    disconnectFromBoolState(sManagerDp + ".Alarms.ValueQueueOverflow", CMWServerBoolState::ValueQueueOverflow, exceptionInfo);
    disconnectFromBoolState(sManagerDp + ".Status.Running", CMWServerBoolState::Running, exceptionInfo);
    dpDisconnect("unCMWServerStatus_HeartBeatTime_CB", sManagerDp + ".Settings.man_Num:_online.._stime");
    dpDisconnect("unCMWServerStatus_HeartBeatPeriod_CB", sManagerDp + ".Settings.man_NumPeriod");
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + sManagerDp + ") FINISH");
  }

  /**
   * Disconnect from one status DPE of type bool. All boolean DPEs use single callback callback function,
   * the argument for dpConnectUserData() is used to identify what DPE was the source of callback.
   *
   *  @param sDpeName The name of DPE to connect to
   *  @param eValueType The userData argument for dpConnectUserData(), will be used in callback processing
   *  @param exceptionInfo Any detected errors will be added to the content of this variable
   *
   *  @return true in case of success   
   */
  private bool disconnectFromBoolState(string sDpeName, CMWServerBoolState eValueType, dyn_string &exceptionInfo)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + sDpeName + ", " + eValueType + ")");
    dpDisconnectUserData("unCMWServerStatus_BoolState_CB", (int)eValueType, sDpeName);
    dyn_errClass err = getLastError();
    if(dynlen(err) > 0)
    {
      fwException_raise(exceptionInfo, "ERROR", __FUNCTION__ + "(): dpDisconnectUserData() failed for " + sDpeName, "");
      DebugFTN(C_DEBUG, __FUNCTION__ + "(" + sDpeName + ", " + eValueType + ") dpDisconnectUserData() failed");
      return false;
    }
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + sDpeName + ", " + eValueType + ") FINISH OK");
    return true;
  }

  /**
   * Process callback data from _Connections.Api.ManNums DPE:n if this manager number in list then
   * the manager must be running and connected to the project.
   *
   *  @param diManNumbers The list of all connected API managers
   */
  private void processManConnection(const dyn_int &diManNumbers)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ")");
    bool bConnected = dynContains(diManNumbers, m_iManNum) > 0;
    if((bConnected != m_bManRunnning) || m_bForceDisplay)  // Can be call right after initialization
    {
      DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") changhe " + m_bManRunnning + " -> " + bConnected);
      m_bManRunnning = bConnected;
      calculateApplyState();
    }
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") FINISH");
  }

  /**
   * Process new heartbeat time from .Settings.man_Num:_online.._stime DPE of configuration
   * DP for this CMWServer manager
   *
   *  @param tHeartBeat The time of heartbeat DPE
   */
  private void processHeartBeatTime(time tHeartBeat)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ")");
    time now = getCurrentTime();
    m_tHeartBeatTime = now > tHeartBeat ? now : tHeartBeat;
    calculateApplyState();
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") FINISH");
  }

  /**
   * Process new state for .Status.Running DPE of configuration DP for this CMWServer manager
   *
   *  @param bState The value of state DPE
   */
  private void processSrvRunning(bool bState)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ")");
    if(bState != m_bCmwRunning)
    {
      DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") changhe " + m_bCmwRunning + " -> " + bState);
      m_bCmwRunning = bState;
      calculateApplyState();
    }
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") FINISH");
  }

  /**
   * Process new state for .Alarms.ValueQueueOverflow DPE of configuration DP for this CMWServer manager
   *
   *  @param bState The value of alarm DPE
   */
  private void processValueQueueOverflow(bool bState)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ")");
    if(bState != m_bValueQueueBad)
    {
      DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") changhe " + m_bValueQueueBad + " -> " + bState);
      m_bValueQueueBad = bState;
      m_bForceDisplay = true;
      calculateApplyState();
    }
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") FINISH");
  }

  /**
   * Process new state for .Alarms.SetRequestQueueOverflow DPE of configuration DP for this CMWServer manager
   *
   *  @param bState The value of alarm DPE
   */
  private void processSetRequestQueueOverflow(bool bState)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ")");
    if(bState != m_bSetReqQueueBad)
    {
      DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") changhe " + m_bSetReqQueueBad + " -> " + bState);
      m_bSetReqQueueBad = bState;
      m_bForceDisplay = true;
      calculateApplyState();
    }
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") FINISH");
  }

  /**
   * Process new state for .Alarms.SetResponseQueueOverflow DPE of configuration DP for this CMWServer manager
   *
   *  @param bState The value of alarm DPE
   */
  private void processSetResponseQueueOverflow(bool bState)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ")");
    if(bState != m_bSetRespQueueBad)
    {
      DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") changhe " + m_bSetRespQueueBad + " -> " + bState);
      m_bSetRespQueueBad = bState;
      m_bForceDisplay = true;
      calculateApplyState();
    }
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") FINISH");
  }

  /**
   * Process new state for .Alarms.RbacDisabled DPE of configuration DP for this CMWServer manager
   *
   *  @param bState The value of alarm DPE
   */
  private void processRbacDisabled(bool bState)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ")");
    if(bState != m_bRbacDisabled)
    {
      DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") changhe " + m_bRbacDisabled + " -> " + bState);
      m_bRbacDisabled = bState;
      m_bForceDisplay = true;
      calculateApplyState();
    }
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") FINISH");
  }

  /**
   * Process new state for .Alarms.DpeConnectionLost DPE of configuration DP for this CMWServer manager
   *
   *  @param bState The value of alarm DPE
   */
  private void processDpeConnectionLost(bool bState)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ")");
    if(bState != m_bDpeConnectionLost)
    {
      DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") changhe " + m_bDpeConnectionLost + " -> " + bState);
      m_bDpeConnectionLost = bState;
      m_bForceDisplay = true;
      calculateApplyState();
    }
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") FINISH");
  }

  /**
   * Process new state for .Alarms.ConfigFileInvalid DPE of configuration DP for this CMWServer manager
   *
   *  @param bState The value of alarm DPE
   */
  private void processConfigFileInvalid(bool bState)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ")");
    if(bState != m_bConfigFileInvalid)
    {
      DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") changhe " + m_bConfigFileInvalid + " -> " + bState);
      m_bConfigFileInvalid = bState;
      m_bForceDisplay = true;
      calculateApplyState();
    }
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") FINISH");
  }

  /**
   * Calculate alarm status, report new alarm state if status has changed (CTRL), or display in table (UI)
   */
  private void calculateApplyState()
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + "), initialized " + m_bInitialized);
    if(!m_bInitialized)
    {
      return;  // Don't calculate new state until initialized
    }
    time tNow = getCurrentTime();
    int iNewState = calculateState(tNow);

    // Apply/display new state
    if((!m_bForceDisplay) && (m_iAlarmState == iNewState))
    {
      return;  // The state is the same as last known state
    }
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") changhe " + m_iAlarmState + " -> " + iNewState );
    m_iAlarmState = iNewState;
    m_tAlarmTime = tNow;
    m_bForceDisplay = false;
    if(myManType() == UI_MAN)
    {
      applyVisibleState();
    }
    else // CTRL
    {
      applyAlarmState();
    }
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") FINISH");
  }

  /**
   * Calculate new alarm state, looks cumbersome but the logic is rather simple
   *
   *  @param tNow The current time
   *
   *  @return New calculated alarm state
   */
  private int calculateState(time tNow)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ")");
    int iResult = (int)CMWServerAlarm::OK;
    if(!m_bManRunnning)  // Is manager running (connected?)
    {
      iResult = (int)CMWServerAlarm::Stopped;
    }
    else // Check if heartbeats arrive in time
    {
      float fIntervalSinceHeartBeat = tNow - m_tHeartBeatTime;
      fIntervalSinceHeartBeat *= 1000;  // sec -> msec
      if((fIntervalSinceHeartBeat * 0.8) > m_iHeartBeatPeriod)  // Allow 20% tolerance
      {
        iResult = (int)CMWServerAlarm::Blocked;
        m_bHeartBeatOk = false;
      }
      else
      {
        m_bHeartBeatOk = true;
      }
    }
    if(iResult == (int)CMWServerAlarm::OK)  // Manager is connected and alive, check further
    {
      if(!m_bCmwRunning)
      {
        iResult = (int)CMWServerAlarm::ServerNotRunning;
      }
      else
      {
        // Combination of individual bits
        if(m_bValueQueueBad)
        {
          iResult |= (1 << (int)CMWServerBoolState::ValueQueueOverflow);
        }
        if(m_bSetReqQueueBad)
        {
          iResult |= (1 << (int)CMWServerBoolState::SetRequestQueueOverflow);
        }
        if(m_bSetRespQueueBad)
        {
          iResult |= (1 << (int)CMWServerBoolState::SetResponseQueueOverflow);
        }
        if(m_bRbacDisabled)
        {
          iResult |= (1 << (int)CMWServerBoolState::RbacDisableIndicator);
        }
        if(m_bDpeConnectionLost)
        {
          iResult |= (1 << (int)CMWServerBoolState::DpeConnectionLost);
        }
        if(m_bConfigFileInvalid)
        {
          iResult |= (1 << (int)CMWServerBoolState::ConfigFileInvalid);
        }
      }
    }
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") FINISH: " + result);
    return iResult;
  }

  /**
   * Display mananger state in table row, shall only be called when running in UI manager
   */
  private void applyVisibleState()
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ")");
    string sTable = "ManagersList";  // TODO: hardcoded??? Column names are also hardcoded
    int iRow = findTableRow(sTable);
    if(iRow < 0)
    {
      DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") Row not found in table");
      return;  // No place to display this manager's info
    }

    // Display of state details is 'hierarchical', for example, if manager is not running
    // then there is no reason of displaying 'lower' level information: state of CMW server, state of queues...
    bool bMayRunCMW = m_bManRunnning && m_bHeartBeatOk;
    // String representation of manager state
    string sManState = m_bManRunnning ? (m_bHeartBeatOk ? "RUN" : "Blocked") : "STOP",
           sManStateColor = m_bManRunnning && m_bHeartBeatOk ? "_3DFace" : "unCMWServer_Bad";
    setMultiValue(
        sTable, "cellBackColRC", iRow, "X", m_iAlarmState == (int)CMWServerAlarm::OK ? "unCMWServer_Ok" : "unCMWServer_Bad",
        sTable, "cellValueRC", iRow, "PVSS", sManState,
        sTable, "cellBackColRC", iRow, "PVSS", sManStateColor,
        sTable, "cellValueRC", iRow, "CMW", bMayRunCMW ? (m_bCmwRunning ? "RUN" : "STOP") : "",
        sTable, "cellBackColRC", iRow, "CMW", (bMayRunCMW && (!m_bCmwRunning)) ? "unCMWServer_Bad" : "_3DFace",
        sTable, "cellValueRC", iRow, "NewValueQ", m_bCmwRunning ? (m_bValueQueueBad ? "Overflow" : "OK") : "",
        sTable, "cellBackColRC", iRow, "NewValueQ", m_bCmwRunning && m_bValueQueueBad ? "unCMWServer_Bad" : "_3DFace",
        sTable, "cellValueRC", iRow, "SetReqQ", m_bCmwRunning ? (m_bSetReqQueueBad ? "Overflow" : "OK") : "",
        sTable, "cellBackColRC", iRow, "SetReqQ", m_bCmwRunning && m_bSetReqQueueBad ? "unCMWServer_Bad" : "_3DFace",
        sTable, "cellValueRC", iRow, "SetRespQ", m_bCmwRunning ? (m_bSetRespQueueBad ? "Overflow" : "OK") : "",
        sTable, "cellBackColRC", iRow, "SetRespQ", m_bCmwRunning && m_bSetRespQueueBad ? "unCMWServer_Bad" : "_3DFace",
        sTable, "cellValueRC", iRow, "Rbac", m_bCmwRunning ? (m_bRbacDisabled ? "Disabled" : "OK") : "",
        sTable, "cellBackColRC", iRow, "Rbac", m_bCmwRunning && m_bRbacDisabled ? "unCMWServer_Bad" : "_3DFace",
        sTable, "cellValueRC", iRow, "DpeLost", m_bCmwRunning ? (m_bDpeConnectionLost ? "LOST" : "OK") : "",
        sTable, "cellBackColRC", iRow, "DpeLost", m_bCmwRunning && m_bDpeConnectionLost ? "unCMWServer_Bad" : "_3DFace",
        sTable, "cellValueRC", iRow, "Config", m_bCmwRunning ? (m_bConfigFileInvalid ? "INVALID" : "OK") : "",
        sTable, "cellBackColRC", iRow, "Config", m_bCmwRunning && m_bConfigFileInvalid ? "unCMWServer_Bad" : "_3DFace",
        sTable, "cellValueRC", iRow, "Time", formatTime("%d-%m-%Y %H:%M:%S", m_tAlarmTime, ".%03d"));
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") FINISH");
  }

  /**
   * Find index of row in table where this manager is shown, the search is done by manager
   * configuration DP name.
   *
   *  @param sTable The name of Table shape where DP name shall be found in "Names" column
   *
   *  @return Index of row in table corresponding to this CMWServer manager; negative value
   *          if not found
   */
  private int findTableRow(string sTable)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ")");
    if(!shapeExists(sTable))
    {
      DebugTN(__FUNCTION__ + ": shape does not exists " + sTable);
      return - 1;
    }
    int iRow, iLineCount;
    getValue(sTable, "lineCount", iLineCount);
    for(iRow = iLineCount - 1 ; iRow >= 0 ; iRow--)
    {
      string sCellContent;
      getValue(sTable, "cellValueRC", iRow, "Names", sCellContent);
      if(sCellContent == m_sManagerDp)
      {
        break;
      }
    }
    if(iRow < 0)
    {
      DebugTN(__FUNCTION__ + ": row for manager " + m_sManagerDp + " not found in " + sTable);
    }
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") FINISH: " + iRow);
    return iRow;
  }

  /**
   * Write calculated alarm state to system ingegrity DP, shall only be called when running in CTRL manager
   */
  private void applyAlarmState()
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + "), state " + m_iAlarmState);
    if((m_sAlarmDpName != "") && dpExists(m_sAlarmDpName + ".alarm"))
    {
      DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + "), DP exists");
      dpSet(m_sAlarmDpName + ".alarm", m_iAlarmState);
    }
    else
    {
      dyn_string exceptionInfo;
      fwException_raise(exceptionInfo, "ERROR", __FUNCTION__ + "(): " + getCatStr("unSystemIntegrity", "WRONG_ALARM_DPE") + 
                        m_sAlarmDpName + ".alarm, value = " + m_iAlarmState, "");
      if(isFunctionDefined("unMessageText_sendException"))
      {
        unMessageText_sendException("*", "*", "SystemIntegrity", "user", "*", exceptionInfo);
      }
    }
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") FINISH");
  }

  /**
   * Check configuration file validity for this manager. It is expected that this
   * function will only be called when running in CTRL manager (system integrity)
   *
   * @param now Current time.
   */
  private void checkConfigFile(time now)
  {
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + "): START");
    m_tConfigFileCheckTime = now;

    bool bFileCorrect = false;

    // Get file name
    string sManagerDp = m_sSystem + m_sManagerDp;
    string sFileName;
    dpGet(sManagerDp + ".Settings.man_ConfigPath", sFileName);
    
    // Check file content
    if(sFileName != "")
    {
      sFileName = getPath(DATA_REL_PATH) + substr(sFileName, strpos(sFileName, "/") + 1);
      mapping mParseErrors;
      string sErrMsg;
      int iTotalItems = CMWServerConfigParser_validateErrorsOnly(sFileName, mParseErrors, sErrMsg);
      bFileCorrect = (sErrMsg == "");
    }
    
    // Write result if changed
    if(bFileCorrect != m_bConfigFileOK)
    {
      m_bConfigFileOK = bFileCorrect;
      dpSet(sManagerDp + ".Alarms.ConfigFileInvalid", ! m_bConfigFileOK);
    }
    DebugFTN(C_DEBUG, __FUNCTION__ + "(" + m_sManagerDp + ") FINISH, new result: " + m_bConfigFileOK);
  }

  /********************* Instance members, one instance == one CMWServer manager in the system ***************************/

  /** @brief Name of DP controlling this manager (without system name) */
  private string m_sManagerDp;
  /** @brief System name, required for correct disconnections */
  private string m_sSystem;
  private int m_iAlarmState = -1;     ///< Calculated alarm state for this manager
  private time m_tAlarmTime;          ///< Time when alarm state was calculated
  private string m_sAlarmDpName;      ///< Name of DP where alarm state shall be reported, only used when run by CTRL manager
  private int m_iManNum;              ///< Manager number in WinCC OA project
  private bool m_bManRunnning;        ///< true if manager is connected to the system, the criteria is: number is found in _Connections.Api.ManNums DPE
  private int m_iHeartBeatPeriod;     ///< Manager must send heartbeats with this period [msec]: .Settings.man_NumPeriod
  private time m_tHeartBeatTime;      ///< Time when last heartbeat was received: .Settings.man_Num:_online.._stime
  private bool m_bHeartBeatOk;        ///< true if heartbeats arrive on time
  private bool m_bCmwRunning;         ///< true if CMW server was started by manager: .Status.Running
  private bool m_bValueQueueBad;      ///< true if queue for new DPE values has overflown: .Alarms.ValueQueueOverflow
  private bool m_bSetReqQueueBad;     ///< true if queue for SET requests has overflown: .Alarms.SetRequestQueueOverflow
  private bool m_bSetRespQueueBad;    ///< true if queue for SET responses has overflown: .Alarms.SetResponseQueueOverflow
  private bool m_bRbacDisabled;       ///< true if RBAC was diabled with command line argument: .Alarms.RbacDisable
  private bool m_bDpeConnectionLost;  ///< true if connection to one or more DPEs was lost while server is running: .Alarms.DpeConnectionLost
  private bool m_bConfigFileInvalid;  ///< true if config file for running server is not valid: .Alarms.ConfigFileInvalid
  private bool m_bInitialized;        ///< true in instance initialization is completed and calculation can proceed
  
  private bool m_bConfigFileOK;       ///< true if result of config file check was successful
  private time m_tConfigFileCheckTime; ///< Time when config file was checked

  /**
   * @brief 'Diaplay' the status even if not changed.
   *
   * The flag used to 'display' the status status even if it is the same as perious status.
   * This flag is used in two situations:
   *  <ol>
   *    <li> if status was never 'displayed' before, i.e. during initialization</li>
   *    <li> in UI - if one of fields has changed, such that overall status will (probably) remain
   *          unmodified, but in UI we also display small details</li>
   *  </ol>
   */
  private bool m_bForceDisplay;
                                      
};

//@)

/** @name Static function: callbacks, thread running */
//@{

/// The function runs in dedicated thread, perodically checks heartbeat arrival for all monitored managers
void unCMWServerStatus_HeartBeatCheck()
{
  while(true)
  {
    CMWServerStatus::checkAllHeartBeats();
    delay(2);  // Reasonable delay?
  }
}

/// Callback for the list of connected API manager numbers
void unCMWServerStatus_AllApiNums_CB(string sDpe, dyn_int diApiNums)
{
  CMWServerStatus::allApiManNumsCallback(diApiNums);
}

/// Callback for heartbeat period of manager
void unCMWServerStatus_HeartBeatPeriod_CB(string sDpe, int iPeriod)
{
  CMWServerStatus::heartBeatPeriodCallback(dpSubStr(sDpe, DPSUB_DP), iPeriod);
}

/// Callback for new heartbeat time of manager
void unCMWServerStatus_HeartBeatTime_CB(string sDpe, time tHeartBeat)
{
  CMWServerStatus::heartBeatTimeCallback(dpSubStr(sDpe, DPSUB_DP), tHeartBeat);
}

/// Callback for bool state DPEs of manager
void unCMWServerStatus_BoolState_CB(int iValueType, string sValueDpe, bool bState)
{
  CMWServerStatus::boolStateCallback(dpSubStr(sValueDpe, DPSUB_DP), iValueType, bState);
}

//@}
