/**@name LIBRARY: unConfigDIP.ctl

@author: Herve Milcent (AB/CO)

Creation Date: 19/04/2005

Modification History:
  29/06/2017: Marcin Bes
    - UCPC-2623: Add DIP front-end to CPC Comm
    This change breaks backwards compatibility of the importation line!
    
  30/05/2011: Herve
    - IS-519: Remove all dependencies on graphical object from the Import device panel in the core import 
    functions, to be able to use the unicos front-end/device check/import functions from a WCCOActrl 
    
  02/05/2011: Herve
  - IS-510: CPC 5 and CPC 6 compatibility of the export/import of the CPC device configuration with all the front-end types
  
  19/02/2009: Herve
    - include unDIP.ctl libs
    
	12/05/2004: Herve
		- DigitalInput: delete archive on evStsReg01 if no archive on PosSt
		
	22/04/2005: Herve
		- implement the event archive: check the event archive name
		- remove UN_CONFIG_DEFAULT_ARCHIVE_SMOOTHING and replace in libs by DPATTR_ARCH_PROC_SIMPLESM
		- allow archive in import file per device
		- add new function DIP_getFrontEndArchiveDp: return the list of Dp that will be archive for the front-end
                
version 1#

External Function :
  . unConfigDIP_checkDeleteCommand : check the coherency of the delete command
  . unConfigDIP_deleteCommand : delete function
  . unConfigDIP_check : check the coherency of the Front-end config
  . unConfigDIP_set : set Front-end config

        
Internal Functions:
        
Purpose:
This library contains the function used to set the DIP.

Usage: Public

PVSS manager usage: Ctrl, NV, NG

Constraints:
  . unGeneration.cat
  . UNICOS-JCOP framework
  . PVSS version: 3.0
  . operating system: NT and W2000, but tested only under W2000.
  . distributed system: yes.
*/

#uses "fwDIP/fwDIP.ctl"

// Constants

// constant for import of UNDIP
const string UNDIP_DPTYPE = "DIP";
const unsigned UN_CONFIG_DIP_LENGTH = 7;
const unsigned UN_CONFIG_UNDIP_COUNTER = 4;
const unsigned UN_CONFIG_UNDIP_DNS = 5;
const unsigned UN_CONFIG_UNDIP_MANAGER_PARAMS = 6;
const unsigned UN_CONFIG_UNDIP_PUBLISHERNAME = 7;

const unsigned DIP_DRIVER = 13;

// constant for UNDIP system integrity
const string c_unSystemIntegrity_UNDIP = "DIP_";

// global variable used during the import to keep the UNDIP data, must be global and of course unique.
global dyn_string g_dsImport_UnDIP_config;
const unsigned UNDIP_COUNTER = 1;
const unsigned UNDIP_DNS = 2;
const unsigned UNDIP_MANAGER_PARAMS = 3;
const unsigned UNDIP_PUBLISHERNAME = 4;

//@{

//------------------------------------------------------------------------------------------------------------------------

// unConfigDIP_checkDeleteCommand
/**
Purpose: check the coherency of the delete command

Parameters:
  dsConfig, dyn_string, input, config dyn_string
  dsDeleteDps, dyn_string, output, list of dp that will be deleted
  exceptionInfo, dyn_string, output, for errors

Usage: External function

PVSS manager usage: Ctrl, NG, NV

Constraints:
	. PVSS version: 3.0
	. operating system: NT and W2000, but tested only under W2000.
	. distributed system: yes.
*/

unConfigDIP_checkDeleteCommand(dyn_string dsConfig, dyn_string &dsDeleteDps, dyn_string &exceptionInfo)
{
	string sFeName, sPLCSubApplication = "", sObjectType = "", sNumber = "", sDpPlcName, sPlcModPlc;
	dyn_string dsAlarmDps;
	
	if (dynlen(dsConfig) >= 2) {				// At least, delete instruction and plc name
		sFeName = dsConfig[2];  // get PLC Name
		if ((sFeName != "") && (dsConfig[1] == UN_DELETE_COMMAND)) {
			if (dynlen(dsConfig) >= 3) {
				sPLCSubApplication = dsConfig[3]; // get Application
			}
			if (dynlen(dsConfig) >= 4) {
				sObjectType = dsConfig[4]; // get DeviceType
			}
			if (dynlen(dsConfig) >= 5) {
				sNumber = dsConfig[5]; // get device number
			}
			unGenericDpFunctions_getPlcDevices("", sFeName, sPLCSubApplication, sObjectType, sNumber, dsDeleteDps);
			// if only delete key word then delete also systemalarm
			if ((sFeName != "") && (sPLCSubApplication == "") && (sObjectType == "") && (sNumber == "")) {
				// sDpPlc is UNDIP_plcName
				sDpPlcName = c_unSystemIntegrity_UNDIP+sFeName;
				if (dpExists(sDpPlcName)) {
					dynAppend(dsDeleteDps, sDpPlcName);
					dsAlarmDps = dpNames(c_unSystemAlarm_dpPattern + "*" + substr(sDpPlcName, strpos(sDpPlcName, ":") + 1),c_unSystemAlarm_dpType);
					dynAppend(dsDeleteDps, dsAlarmDps);
				}
			}
		}
		else {
			fwException_raise(exceptionInfo,"ERROR","unConfigDIP_checkDeleteCommand:" + getCatStr("unGeneration","BADINPUTS"),"");
		}
	}
	else {
		fwException_raise(exceptionInfo,"ERROR","unConfigDIP_checkDeleteCommand:" + getCatStr("unGeneration","BADINPUTS"),"");
	}
}

//------------------------------------------------------------------------------------------------------------------------
//unConfigDIP_check
/**
Purpose: check the coherency of the DIP front-end config

Parameters:
  currentObject: string, input, type of config
  currentLineSplit: dyn_string, input, config
  dsDeleteDps: dyn_string, input, dp to be deleted
  exceptionInfo: dyn_string, output, for errors

Usage: External function

PVSS manager usage: Ctrl, NG, NV

Constraints:
        . PVSS version: 3.0
        . operating system: NT and W2000, but tested only under W2000.
        . distributed system: yes.
*/
unConfigDIP_check(string currentObject, dyn_string currentLineSplit, dyn_string dsDeleteDps, dyn_string &exceptionInfo)
{
	dyn_dyn_string ddsComponentInfo;
	bool bDIPInstalled=false;

	// get the list of component installed
	fwInstallation_getInstalledComponents(ddsComponentInfo);

	for(int i=1;i<=dynlen(ddsComponentInfo); i++) {
		if(ddsComponentInfo[i][1] == "fwDIP") {
			bDIPInstalled = true;
        }
	}
	if(bDIPInstalled) {
		switch(currentObject) {
			case UN_PLC_COMMAND:
				unConfigDIP_checkApplication(currentLineSplit, dsDeleteDps, exceptionInfo);
				break;
			default:
	 			fwException_raise(exceptionInfo, "ERROR", getCatStr("unGeneration","UNKNOWNFUNCTION"), "");
				break;
		}
	}
	else {
	 	fwException_raise(exceptionInfo, "ERROR", "fwDIP not installed", "");
    }
}

//------------------------------------------------------------------------------------------------------------------------

// unConfigDIP_deleteCommand
/**
Purpose: delete function

Parameters:
  dsConfig, dyn_string, input, config dyn_string
  exceptionInfo, dyn_string, output, for errors

Usage: External function

PVSS manager usage: Ctrl, NG, NV

Constraints:
	. PVSS version: 3.0
	. operating system: NT and W2000, but tested only under W2000.
	. distributed system: yes.

@reviewed 2018-07-25 @whitelisted{Callback}

*/

unConfigDIP_deleteCommand(dyn_string dsConfig, dyn_string &exceptionInfo)
{
	string sPLCSubApplication = "", sObjectType = "", sNumber = "";
	dyn_string dsDpeList, exceptionInfoTemp;
	
	unConfigDIP_checkDeleteCommand(dsConfig, dsDpeList, exceptionInfoTemp);
	if (dynlen(exceptionInfoTemp) > 0)
	{
		dynAppend(exceptionInfo, exceptionInfoTemp);
		fwException_raise(exceptionInfo,"ERROR","unConfigDIP_deleteCommand:" + getCatStr("unGeneration","BADINPUTS"),"");
	}
	else
	{
		if (dynlen(dsConfig) >= 2)				// At least, delete instruction and plc name
		{
			if (dynlen(dsConfig) >= 3)
			{
				sPLCSubApplication = dsConfig[3];
			}
			if (dynlen(dsConfig) >= 4)
			{
				sObjectType = dsConfig[4];
			}
			if (dynlen(dsConfig) >= 5)
			{
				sNumber = dsConfig[5];
			}
// delete first all the subscriptions and publications before deleting the devices
			unConfigDIP_deleteSubscription(dsDpeList, exceptionInfo);
            unConfigDIP_deletePublication(dsDpeList, exceptionInfo, "DIPConfig_" + dsConfig[2]);
				
			unGenericDpFunctions_deletePlcDevices(dsConfig[2], sPLCSubApplication, sObjectType, sNumber, "progressBar", exceptionInfo);
			if ((dsConfig[2] != "") && (sPLCSubApplication == "") && (sObjectType == "") && (sNumber == "")) {
				unConfigDIP_deleteFrontEnd(dsConfig[2], exceptionInfo);
			}
		}
	}
}

//------------------------------------------------------------------------------------------------------------------------
//unConfigDIP_deleteSubscription
/**
Purpose: delete the DIP subscriptions

Parameters:
  dsDp: dyn_string, input, list of the dpes for which subscriprions should be deleted
  exceptionInfo: dyn_string, output, for errors

Usage: External function

PVSS manager usage: Ctrl, NG, NV

Constraints:
	. PVSS version: 3.0
	. operating system: NT and W2000, but tested only under W2000.
	. distributed system: yes.
*/
unConfigDIP_deleteSubscription(dyn_string dsDp, dyn_string &exceptionInfo)
{
	for(int i = 1; i <= dynlen(dsDp); i++) {
        string dpName = dsDp[i];
        string deviceType = dpTypeName(dpName);
        
        if(strtolower(substr(deviceType,0,3)) == "cpc") // handling CPC objects
        {
            mapping addresses;  // set of device type's addresses
            evalScript(addresses, "mapping main() { return CPC_" + cpcConfigGenericFunctions_getShortTypeName(deviceType) + "Config_getConfig(); }", makeDynString()); 
            dyn_string keys = mappingKeys(addresses);
            for (int j = 1; j <= dynlen(keys); j++) {
                string key = keys[j];
                string address = addresses[key]["address"];
                // add all Input signals (potentially subscribed)
                if(strpos(address, "ProcessInput") >= 0) {
                    fwDIP_unsubscribe(dpName + address, exceptionInfo);
                }
            }
            
        }
        // here put the logic for unsubscribing other objects (was never done)
	}

// does not work like that
//	_fwDIP_removeSubscription(dsDpe, exceptionInfo);
}

//------------------------------------------------------------------------------------------------------------------------
//unConfigDIP_deletePublication
/**
Purpose: delete the DIP publications

Parameters:
  dsDp: dyn_string, input, list of dpes for which publications should be deleted
  exceptionInfo: dyn_string, output, for errors
  configDp: string, input, name of the DIPConfig_ dpe

Usage: External function

PVSS manager usage: Ctrl, NG, NV

Constraints:
	. PVSS version: 3.0
	. operating system: NT and W2000, but tested only under W2000.
	. distributed system: yes.
*/
unConfigDIP_deletePublication(dyn_string dsDp, dyn_string &exceptionInfo, string configDp="")
{
    dyn_dyn_string publicationConfig;
    fwDIP_getAllPublications(configDp, publicationConfig, exceptionInfo);
    if(dynlen(publicationConfig) >= fwDIP_OBJECT_DPES) { // unpublishing makes sense only if the config is not empty; UNDEVICE-27
        for(int i=1;i<=dynlen(dsDp);i++) {
            string deviceName = dsDp[i];
            string deviceType = dpTypeName(deviceName);
        
            if(strtolower(substr(deviceType,0,3)) == "cpc") { // handling CPC objetcs
                mapping addresses;
                evalScript(addresses, "mapping main() { return CPC_" + cpcConfigGenericFunctions_getShortTypeName(deviceType) + "Config_getConfig(); }", makeDynString()); 
                dyn_string keys = mappingKeys(addresses);
                for (int j = 1; j <= dynlen(keys); j++) {
                    string key = keys[j];
                    string address = addresses[key]["address"];
                    // check each output signal (potentially published)
                    if(strpos(address, "ProcessOutput") >= 0) {
                        // check if dpe is published
                        int iResult = dynContains(publicationConfig[fwDIP_OBJECT_DPES], deviceName + address);
                        if(iResult > 0) {
                            string dipItem = publicationConfig[fwDIP_OBJECT_ITEM][iResult];
                            fwDIP_unpublish(dipItem, exceptionInfo);
                        }
                    }
                }
            }
        }
        // here implement logic for unpublishing other objects (was never done)
	}
}

//------------------------------------------------------------------------------------------------------------------------
//unConfigDIP_set
/**
Purpose: set DIP front-end config

Parameters:
  currentObject: string, input, type of config
  currentLineSplit: dyn_string, input, config
  exceptionInfo: dyn_string, output, for errors

Usage: External function

PVSS manager usage: Ctrl, NG, NV

Constraints:
	. PVSS version: 3.0
	. operating system: NT and W2000, but tested only under W2000.
	. distributed system: yes.
*/
unConfigDIP_set(string currentObject, dyn_string currentLineSplit, dyn_string &exceptionInfo)
{
	dyn_dyn_string ddsComponentInfo;
	bool bDIPInstalled=false;

	// get the list of component installed
	fwInstallation_getInstalledComponents(ddsComponentInfo);

	for(int i=1;i<=dynlen(ddsComponentInfo); i++) {
		if(ddsComponentInfo[i][1] == "fwDIP") {
			bDIPInstalled = true;
        }
	}
	if(bDIPInstalled) {
		switch(currentObject) {
			case UN_PLC_COMMAND:
				g_dsImport_UnDIP_config = makeDynString();
				unConfigDIP_setApplication(currentLineSplit, exceptionInfo);
                // TODO: check if code below is used
				if(dynlen(exceptionInfo) <= 0) {
					g_dsImport_UnDIP_config[UNDIP_COUNTER] = currentLineSplit[UN_CONFIG_UNDIP_COUNTER];
                    g_dsImport_UnDIP_config[UNDIP_DNS] = currentLineSplit[UN_CONFIG_UNDIP_DNS];
                    g_dsImport_UnDIP_config[UNDIP_MANAGER_PARAMS] = currentLineSplit[UN_CONFIG_UNDIP_MANAGER_PARAMS];
                    g_dsImport_UnDIP_config[UNDIP_PUBLISHERNAME] = currentLineSplit[UN_CONFIG_UNDIP_PUBLISHERNAME];

				}
				break;
			default:
	 			fwException_raise(exceptionInfo, "ERROR", getCatStr("unGeneration","UNKNOWNFUNCTION"), "");
				break;
		}
	}
	else
	 	fwException_raise(exceptionInfo, "ERROR", "fwDIP not installed", "");
}

//------------------------------------------------------------------------------------------------------------------------

// unConfigDIP_checkApplication
/**
Purpose: check data for a sub application (not comm)

Parameters:
		dsPLC: dyn_string, input, DIP configuration line in file + additional parameters
			dsPLC must contains :
			1. string, DIP
			2. string, DIP FrontEnd name
			3. string, DIP subapplication name
			4. string, Publisher Name
            5. string, Counter address 
            6. string, comma-separate list of DIP dns servers to be used by the DIP manager
			(additional parameters)
			5. driver number 
			6. boolean archive name
			7. analog archive name
            8. event = 16/32 bit
            9. event archive name
		dsDeleteDps, dyn_string, input, datapoints that will be deleted before importation
		exceptionInfo: dyn_string, output, for errors

Usage: External function

PVSS manager usage: Ctrl, NG, NV

Constraints:
	. PVSS version: 3.0 
	. operating system: NT and W2000, but tested only under W2000.
	. distributed system: yes.
*/

unConfigDIP_checkApplication(dyn_string dsPLC, dyn_string dsDeleteDps, dyn_string &exceptionInfo)
{
	dyn_string dsPLCLine, dsPLCAdditionalData;
	int i, iRes, iDriverNum;
	string sPLCDp, sHostname, sAliasDp, tempDp;

	if (dynlen(dsPLC) == (UN_CONFIG_DIP_LENGTH + UN_CONFIG_PLC_ADDITIONAL_LENGTH)) {
		dsPLCLine = dsPLC;
		for(i=1;i<=UN_CONFIG_PLC_ADDITIONAL_LENGTH;i++) {
			dynAppend(dsPLCAdditionalData, dsPLCLine[UN_CONFIG_DIP_LENGTH+1]);
			dynRemove(dsPLCLine,UN_CONFIG_DIP_LENGTH+1);
        }
		unConfigDIP_checkData(dsPLCLine, exceptionInfo);
		unConfigDIP_checkAdditionalData(dsPLCAdditionalData, exceptionInfo);

        // first check if the PLC name is correct:
		tempDp = dsPLCLine[UN_CONFIG_PLC_NAME];
		if(!unConfigGenericFunctions_nameCheck(tempDp)) {
			fwException_raise(exceptionInfo,"ERROR","unConfigDIP_checkApplication: " + dsPLCLine[UN_CONFIG_PLC_NAME] + getCatStr("unGeneration","BADPLCNAME"),"");
			return;
		}

        // check if the PLC Application Name is correct:
		tempDp = dsPLCLine[UN_CONFIG_PLC_SUBAPPLICATION];
		if(!unConfigGenericFunctions_nameCheck(tempDp)) {
			fwException_raise(exceptionInfo,"ERROR","unConfigDIP_checkApplication: " + dsPLCLine[UN_CONFIG_PLC_SUBAPPLICATION] + getCatStr("unGeneration","BADPLCAPPLICATIONNAME"),"");
			return;
		}

        // DIP-specific
        // check if _FwDPIConfig exists for this Front-End
        string sDIPConfigDP = "DIPConfig_" + dsPLCLine[UN_CONFIG_PLC_NAME];
        if(!dpExists(sDIPConfigDP)) {
            unImportDevice_report(0, "WARNING: DIP Configuration " + sDIPConfigDP + " does not exist. This config and new DIP manager will be created during importation.");
        }
        else {
            dyn_int diNumber;
            dyn_string dsConfigFileDIPConfig;
            // get all DIP configuration info from config file
            int iResult = _fwDIP_getAllDIPConfigFromConfigFile(diNumber, dsConfigFileDIPConfig);
            if(iResult != 0) {
                unImportDevice_report(0, "WARNING: Cannot retrieve DIP config info from config file.");
            }
            else {    // DIP config retrieved
                // check if DIPConfig_FrontEndName exists
                iResult = dynContains(dsConfigFileDIPConfig, sDIPConfigDP);
                if (iResult == 0) {
                    unImportDevice_report(0, "WARNING: DIP Configuration " + sDIPConfigDP + " is not in the config file. It will be added during importation.");
                }
                else {
                    int iDIPManagerNum = diNumber[iResult]; // get associated DIP manager number ([dip_x])
                    
                    string options;
                    if (iDIPManagerNum != 1) {
                        options = "-num " + iDIPManagerNum;
                    }
                    unImportDevice_report(0, "INFO: DIP Configuration " + sDIPConfigDP + " exists in the config file fpr DIP Manager " + iDIPManagerNum + " and it will be used for importation.");
                    
                    string sHost, sIP;
                    int iPort;
                    unGenericDpFunctions_getHostName(getSystemName(), sHost, sIP);
                    unGenericDpFunctions_getpmonPort(getSystemName(), iPort);
                    
                    dyn_mixed managerInfo;
                    fwInstallationManager_getProperties("WCCOAdip", options, managerInfo, sHost, iPort);  // check if manager x already exists in the project
                    if(managerInfo[FW_INSTALLATION_MANAGER_PMON_IDX] == -1) {
                        unImportDevice_report(0, "WARNING: DIP Manager number " + iDIPManagerNum + "(options: " + options + ") does not exist. It will be added to the project during importation.");
                    }
                }
            }
        }
        

        // sPLCDp is UNDIP_plcName
		sPLCDp = getSystemName() +  c_unSystemIntegrity_UNDIP + dsPLCLine[UN_CONFIG_PLC_NAME];
        // sAliasDp = dpAliasToName(dsPLCLine[UN_CONFIG_PLC_NAME]);
        sAliasDp = unGenericDpFunctions_dpAliasToName(dsPLCLine[UN_CONFIG_PLC_NAME]);

		if (substr(sAliasDp, strlen(sAliasDp) - 1) == ".") {
			sAliasDp = substr(sAliasDp, 0, strlen(sAliasDp) - 1);
		}
		if ((sAliasDp != "") && (sAliasDp != sPLCDp)) {
			if (dynContains(dsDeleteDps, sAliasDp) <= 0) {		// Alias exists, is not used by the "DIP" config datapoint and will not be deleted
				fwException_raise(exceptionInfo,"ERROR","unConfigDIP_checkApplication: " + dsPLCLine[UN_CONFIG_PLC_NAME] + getCatStr("unGeneration","BADHOSTNAMEALIAS") + sAliasDp,"");
			}
		}
		if (dpExists(sPLCDp) && (dynContains(dsDeleteDps, sPLCDp) <= 0)) {
			sHostname = unGenericDpFunctions_getAlias(sPLCDp);
			iRes = dpGet(sPLCDp + ".communication.driver_num", iDriverNum);
			if (iRes >= 0) {
				if (iDriverNum != dsPLCAdditionalData[UN_CONFIG_PLC_ADDITIONAL_DRIVER_NUM]) {
					fwException_raise(exceptionInfo,"ERROR","unConfigDIP_checkApplication: " + getCatStr("unGeneration","BADEXISTINGPLCDRIVER")+iDriverNum,"");
				}
				if (sHostname != dsPLCLine[UN_CONFIG_PLC_NAME]) {
					fwException_raise(exceptionInfo,"ERROR","unConfigDIP_checkApplication: " + getCatStr("unGeneration","BADEXISTINGPLCHOSTNAME"),"");
				}
			}
		}
	}
	else {
		fwException_raise(exceptionInfo,"ERROR","unConfigDIP_checkApplication: " + getCatStr("unGeneration","BADINPUTS"),"");
	}
}

//------------------------------------------------------------------------------------------------------------------------

// unConfigDIP_checkData
/**
Purpose: check PLC data for configuration

Parameters:
		dsPLC: dyn_string, input, DIP configuration line in file + additional parameters
			dsPLC must contains :
			1. string, DIP
			2. string, DIP FrontEnd name
			3. string, DIP subapplication name
			4. string, Publisher Name
            5. string, Counter address 
            6. string, comma-separate list of DIP dns servers to be used by the DIP manager
			(additional parameters)
			5. driver number 
			6. boolean archive name
			7. analog archive name
            8. event = 16/32 bit
            9. event archive name
		dsDeleteDps, dyn_string, input, datapoints that will be deleted before importation
		exceptionInfo: dyn_string, output, for errors

Usage: External function

PVSS manager usage: Ctrl, NG, NV

Constraints:
	. PVSS version: 3.0 
	. operating system: NT and W2000, but tested only under W2000.
	. distributed system: yes.
*/

unConfigDIP_checkData(dyn_string dsPLC, dyn_string &exceptionInfo)
{
	unsigned uPLCNumber;
	
	if (dynlen(dsPLC) != UN_CONFIG_DIP_LENGTH) {
		fwException_raise(exceptionInfo,"ERROR","unConfigDIP_checkData:" + getCatStr("unGeneration","BADINPUTS"),"");
	}
	else {
		unConfigGenericFunctions_checkPLCType(dsPLC[UN_CONFIG_PLC_TYPE], exceptionInfo);

		if (strrtrim(dsPLC[UN_CONFIG_PLC_NAME]) == "") {
			fwException_raise(exceptionInfo,"ERROR","unConfigDIP_checkData:" + getCatStr("unGeneration","ERRORPLCNAME"),"");
		}
		
		if (strrtrim(dsPLC[UN_CONFIG_PLC_SUBAPPLICATION]) == "") {
			fwException_raise(exceptionInfo,"ERROR","unConfigDIP_checkData:" + getCatStr("unGeneration","ERRORSUBAPPLI"),"");
		}			

	}
}

//------------------------------------------------------------------------------------------------------------------------

// unConfigDIP_checkAdditionalData
/**
Purpose: check additional data for configuration

Parameters:
		dsPLC: dyn_string, input, DIP configuration line in file + additional parameters
			dsPLC must contains :
			1. string, DIP
			2. string, DIP FrontEnd name
			3. string, DIP subapplication name
			4. string, Publisher Name
            5. string, Counter address 
            6. string, comma-separate list of DIP dns servers to be used by the DIP manager
			(additional parameters)
			5. driver number 
			6. boolean archive name
			7. analog archive name
            8. event = 16/32 bit
            9. event archive name
		exceptionInfo: dyn_string, output, for errors

Usage: External function

PVSS manager usage: Ctrl, NG, NV

Constraints:
	. PVSS version: 3.0 
	. operating system: NT and W2000, but tested only under W2000.
	. distributed system: yes.
*/

unConfigDIP_checkAdditionalData(dyn_string dsPLC, dyn_string &exceptionInfo)
{
	if (dynlen(dsPLC) != UN_CONFIG_PLC_ADDITIONAL_LENGTH) {
		fwException_raise(exceptionInfo,"ERROR","unConfigDIP_checkAdditionalData:" + getCatStr("unGeneration","BADINPUTS"),"");
	}
	else {
		unsigned uPLCDriver = (unsigned)dsPLC[UN_CONFIG_PLC_ADDITIONAL_DRIVER_NUM];
		if(uPLCDriver != DIP_DRIVER)  {
			fwException_raise(exceptionInfo,"ERROR","unConfigDIP_checkAdditionalData: driver 13 must be used","");
		}
		if (dsPLC[UN_CONFIG_PLC_ADDITIONAL_ARCHIVE_BOOL] == "") {
			fwException_raise(exceptionInfo,"ERROR","unConfigDIP_checkAdditionalData: bool archive" + getCatStr("unGeneration","ERRORARCHIVEEMPTY"),"");
		}
		if (dsPLC[UN_CONFIG_PLC_ADDITIONAL_ARCHIVE_ANALOG] == "") {
			fwException_raise(exceptionInfo,"ERROR","unConfigDIP_checkAdditionalData: analog archive" + getCatStr("unGeneration","ERRORARCHIVEEMPTY"),"");
		}
		if (dsPLC[UN_CONFIG_PLC_ADDITIONAL_ARCHIVE_EVENT] == "") {
			fwException_raise(exceptionInfo,"ERROR","unConfigDIP_checkAdditionalData: event archive" + getCatStr("unGeneration","ERRORARCHIVEEMPTY"),"");
		}
	}
}

//------------------------------------------------------------------------------------------------------------------------
// unConfigDIP_setApplication
/**
Purpose: set PLC configuration for a sub application (not comm)

Parameters:
		dsPLC: dyn_string, input, DIP configuration line in file + additional parameters
			dsPLC must contains :
			1. string, DIP
			2. string, DIP FrontEnd name
			3. string, DIP subapplication name
			4. string, Publisher Name
            5. string, Counter address 
            6. string, comma-separate list of DIP dns servers to be used by the DIP manager
			(additional parameters)
			5. driver number 
			6. boolean archive name
			7. analog archive name
            8. event = 16/32 bit
            9. event archive name
		exceptionInfo: dyn_string, output, for errors

Usage: External function

PVSS manager usage: Ctrl, NG, NV

Constraints:
	. First line of a generation file must contain the PLC description. Fields are defined in constants at this library beginning
	. PVSS version: 3.0 
	. operating system: NT and W2000, but tested only under W2000.
	. distributed system: yes.
*/
unConfigDIP_setApplication(dyn_string dsPLCData, dyn_string &exceptionInfo)
{
	int iPLCNumber, iRes, i, position;
	string sDpName, addressReference, tempDp;
	dyn_string dsPLC, dsAdditionalParameters, dsSubApplications, dsImportTimes, dsBoolArchives, dsAnalogArchives, exceptionInfoTemp;
	dyn_string dsEventArchives;
	
	// Check
	
	if (dynlen(dsPLCData) == (UN_CONFIG_DIP_LENGTH + UN_CONFIG_PLC_ADDITIONAL_LENGTH)) {
	// 1. Separate config file info and additional parameters
		dsPLC = dsPLCData;
		dynClear(dsAdditionalParameters);
		for(i=1;i<=UN_CONFIG_PLC_ADDITIONAL_LENGTH;i++) {
			dynAppend(dsAdditionalParameters, dsPLC[UN_CONFIG_DIP_LENGTH+1]);
			dynRemove(dsPLC,UN_CONFIG_DIP_LENGTH+1);
		}
	
	// check if the PLC Application Name is correct:
		tempDp = dsPLCData[UN_CONFIG_PLC_SUBAPPLICATION];
		if(!unConfigGenericFunctions_nameCheck(tempDp)) {
			fwException_raise(exceptionInfo,"ERROR","unConfigDIP_setApplication: " + dsPLCData[UN_CONFIG_PLC_SUBAPPLICATION] + getCatStr("unGeneration","BADPLCAPPLICATIONNAME"),"");
			return;
		}
        
	// Check if fwDIPConfig for this FrontEnd already exists
        string sFwDIPConfigDP = "DIPConfig_" + dsPLCData[UN_CONFIG_PLC_NAME]; // _FwDIPConfig name for FrontEnd
        if(!dpExists(sFwDIPConfigDP)) {
            iRes = dpCreate(sFwDIPConfigDP, "_FwDipConfig");   // try to create
            if(iRes != 0) {
                fwException_raise(exceptionInfo,"ERROR","unConfigDIP_setApplication: unable to create DIP Configuration " + sDIPConfigDP + getCatStr("unGeneration","DPCREATEFAILED"),"");
                return;
            }
            // add publisher name
            iRes = dpSet(sFwDIPConfigDP + ".publishName", dsPLCData[UN_CONFIG_UNDIP_PUBLISHERNAME]);
            if(iRes != 0)  {
                fwException_raise(exceptionInfo,"ERROR","unConfigDIP_setApplication: unable to set DIP publisher name (" + dsPLCData[UN_CONFIG_UNDIP_PUBLISHERNAME] + ") in config " + sDIPConfigDP + ".publishName " + getCatStr("unGeneration","DPCREATEFAILED"),"");
                return;
            }
            unImportDevice_report(0, "INFO: DIP Configuration " + sFwDIPConfigDP + " was created.");
        }
        
    // Check if [dip_x] entry for this DIPConfig exists in the config file
        dyn_int diNumber;
        dyn_string dsConfigFileDIPConfig;
        int iResult = _fwDIP_getAllDIPConfigFromConfigFile(diNumber, dsConfigFileDIPConfig);    // retrieve all the DIP Configs from config file
        if(iResult != 0) {
            fwException_raise(exceptionInfo,"ERROR","Cannot retrieve DIP config info from config file.","");
            return;
        }
        iResult = dynContains(dsConfigFileDIPConfig, sFwDIPConfigDP); // check if _FwDIPConfig_FEName is already in use inside config
        int iDIPManagerNum;
        
        string sHost, sIP;
        int iPort;

        unGenericDpFunctions_getHostName(getSystemName(), sHost, sIP);
        unGenericDpFunctions_getpmonPort(getSystemName(), iPort);
        
        if(iResult == -1) {
            fwException_raise(exceptionInfo,"ERROR","Critical problem in dynContains function.","");
            return;
        } 
        else if(iResult != 0) {   // _FwDIPConfig_FEName is in the config
            iDIPManagerNum = diNumber[iResult]; // read the DIP manager number
            unImportDevice_report(0, "INFO: DIP Configuration " + sFwDIPConfigDP + " exists in the config for DIP manager number " + iDIPManagerNum);
        }
        else {   // _FwDIPConfig_FEName is not in the config
            // find first free DIP manager number - meaning no section [dip_x] AND not Manager with param -num x
            iDIPManagerNum = 2; // skip manager number 1, as it's DIP default one and it can be without -num 1 cfg
            dyn_mixed managerInfo;
            string options = "-num 2";
            while(dynContains(diNumber, iDIPManagerNum) || (fwInstallationManager_getProperties("WCCOAdip", options, managerInfo, sHost, iPort)==0 && managerInfo[FW_INSTALLATION_MANAGER_PMON_IDX] != -1))
            {
                iDIPManagerNum++;
                options = "-num " + iDIPManagerNum;
            }
            string configPath = getPath(CONFIG_REL_PATH, "config");
            // insert [dip_x] section in config file
            if(paCfgInsertValue(configPath, "dip_" + iDIPManagerNum, "DipConfig", sFwDIPConfigDP) != 0) {
                fwException_raise(exceptionInfo,"ERROR","Cannot create DIP Configuration [dip_" + iDIPManagerNum + "] in the config file " + configPath,"");
                return;
            }
            unImportDevice_report(0, "INFO: [dip_" + iDIPManagerNum + "] section was added to the config file (" + configPath + ") with config: DipConfig = \"" + sFwDIPConfigDP + "\"");
        }
        // check if DIP manager number iDIPManagerNum exists
        dyn_mixed managerInfo;
        string options = "-num " + iDIPManagerNum;
        fwInstallationManager_getProperties("WCCOAdip", options, managerInfo, sHost, iPort);
        if(managerInfo[FW_INSTALLATION_MANAGER_PMON_IDX] == -1) { // manager does not exist
            // add manager
            fwInstallationManager_add("WCCOAdip", "always", 30, 2, 2, "-num " + iDIPManagerNum + " -dns \"" + dsPLCData[UN_CONFIG_UNDIP_DNS] + "\" " + dsPLCData[UN_CONFIG_UNDIP_MANAGER_PARAMS], sHost, iPort);
            unImportDevice_report(0, "INFO: added DIP manager number " + iDIPManagerNum + " to the project.");
            
            // start manager
            if(fwInstallationManager_command("START", "WCCOAdip", options, sHost, iPort) == -1) {
                unImportDevice_report(0, "WARNING: Cannot start DIP Manager number " + iDIPManagerNum + ". Must be started before using this DIP Front-End.");
            }
            else {
                unImportDevice_report(0, "INFO: DIP Manager number " + iDIPManagerNum + " started.");
            }
        }
    
        // 3. Set PLC configuration in UnDIP
		sDpName = c_unSystemIntegrity_UNDIP + dsPLC[UN_CONFIG_PLC_NAME];
		tempDp = sDpName;
		unConfigGenericFunctions_createDp(sDpName, UNDIP_DPTYPE, exceptionInfoTemp);

		if (dynlen(exceptionInfoTemp) <= 0) {
			sDpName = getSystemName() + sDpName;
			unConfigGenericFunctions_setAlias(sDpName, dsPLC[UN_CONFIG_PLC_NAME], exceptionInfo);
			iRes = dpGet(sDpName + ".configuration.subApplications", dsSubApplications,
						 sDpName + ".configuration.importTime", dsImportTimes,
						 sDpName + ".configuration.archive_bool", dsBoolArchives,
						 sDpName + ".configuration.archive_analog", dsAnalogArchives,
						 sDpName + ".configuration.archive_event", dsEventArchives);
			position = dynContains(dsSubApplications, dsPLC[UN_CONFIG_PLC_SUBAPPLICATION]);
			if (position <= 0) {
				position = dynlen(dsSubApplications) + 1;
			}
            
            // if the event archive is empty, set it to bool archive
			if(dynlen(dsEventArchives) <= 0) {
				dsEventArchives = dsBoolArchives;
            }
				
			dsSubApplications[position] = dsPLC[UN_CONFIG_PLC_SUBAPPLICATION];
			dsImportTimes[position] = (string)getCurrentTime();
			dsBoolArchives[position] = dsAdditionalParameters[UN_CONFIG_PLC_ADDITIONAL_ARCHIVE_BOOL];
			dsAnalogArchives[position] = dsAdditionalParameters[UN_CONFIG_PLC_ADDITIONAL_ARCHIVE_ANALOG];
			dsEventArchives[position] = dsAdditionalParameters[UN_CONFIG_PLC_ADDITIONAL_ARCHIVE_EVENT];
            
            iRes = dpSetWait(sDpName + ".configuration.importTime", dsImportTimes,
							 sDpName + ".configuration.archive_bool", dsBoolArchives,
							 sDpName + ".configuration.archive_analog", dsAnalogArchives,
							 sDpName + ".configuration.archive_event", dsEventArchives,
                             sDpName + ".configuration.subApplications",dsSubApplications,
							 sDpName + ".communication.driver_num", dsAdditionalParameters[UN_CONFIG_PLC_ADDITIONAL_DRIVER_NUM]);
			if (iRes < 0) {
				fwException_raise(exceptionInfo,"ERROR","unConfigDIP_setApplication: " + getCatStr("unPVSS","DPSETFAILED"),"");
			}	

            // 4. Create _UnSystemAlarm datapoints
			unSystemIntegrity_createSystemAlarm(tempDp,
                                                DS_pattern,
                                                getCatStr("unSystemIntegrity", "PLC_DS_DESCRIPTION") + dsPLC[UN_CONFIG_PLC_NAME] + " -> DS DIP manager " + iDIPManagerNum,
                                                exceptionInfo);
		}
		else {
			dynAppend(exceptionInfo, exceptionInfoTemp);
		}
		
		// set counter
		unConfigDIP_addSubscription(sDpName + ".communication.counter", dsPLC[UN_CONFIG_PLC_NAME], dsPLC[UN_CONFIG_UNDIP_COUNTER], exceptionInfo);
		
	}
	else {
		fwException_raise(exceptionInfo,"ERROR","unConfigDIP_setApplication: " + getCatStr("unGeneration","BADINPUTS"),"");
	}
}

//---------------------------------------------------------------------------------------------------------------------------------------

// unConfigDIP_deleteFrontEnd
/**
Purpose: delete Front-end config

Parameters : 
	sPlcName, string, input, plc name
	exceptionInfo, dyn_string, output, for errors
	
Usage: Public

PVSS manager usage: Ctrl, UI (WCCOAui)

Constraints:
	. PVSS version: 3.0 
	. operating system: NT and W2000, but tested only under W2000.
	. distributed system: yes.
*/

unConfigDIP_deleteFrontEnd(string sPlcName, dyn_string& exceptionInfo)
{
	string sDpPlcName, sPlcModPlc="", exceptionText;
	int iRes;
	dyn_string dsAlarmDps;

	sDpPlcName = c_unSystemIntegrity_UNDIP+sPlcName;
	if (!dpExists(sDpPlcName)) {
        sDpPlcName = "";
    }

	if (sDpPlcName != "") {
		if (unGenericDpFunctions_getSystemName(sDpPlcName) == getSystemName()) {
            // 2. Delete _Un_Plc
            string unDIPdp = substr(sDpPlcName, strpos(sDpPlcName, ":") + 1);
            
            // unsubscribe counter before its dp will be deleted 
            fwDIP_unsubscribe(unDIPdp + ".communication.counter", exceptionInfo);
    
			iRes = dpDelete(unDIPdp);
			if (iRes < 0) {
				fwException_raise(exceptionInfo, "ERROR", "unConfigDIP_deleteFrontEnd: " + unDIPdp + getCatStr("unGeneral", "UNOBJECTNOTDELETED"), "");
			}
            // 3. Delete alarm system 
			dsAlarmDps = dpNames(c_unSystemAlarm_dpPattern + "*" + unDIPdp,c_unSystemAlarm_dpType);
			for(int i=1;i<=dynlen(dsAlarmDps);i++) {
				_unSystemIntegrity_modifyApplicationAlertList(dsAlarmDps[i], false, exceptionInfo);
				iRes = dpDelete(dsAlarmDps[i]);
				if (iRes < 0) {
					fwException_raise(exceptionInfo, "ERROR", "unConfigDIP_deleteFrontEnd: " + dsAlarmDps[i] + getCatStr("unGeneral", "UNOBJECTNOTDELETED"), "");
				}
			}
		}
		else {
			fwException_raise(exceptionInfo, "ERROR", "unConfigDIP_deleteFrontEnd: " + sDpPlcName + getCatStr("unGeneral", "UNOBJECTNOTDELETEDBADSYSTEM"), "");
		}
	}
    
    // 4. if no subscriptions and publications left in the config, delete config+manager
    string sFwDIPConfigDP = "DIPConfig_" + sPlcName;
    
    dyn_string clientConfig, serverConfig;
    dpGet(sFwDIPConfigDP + ".clientConfig", clientConfig);
    dpGet(sFwDIPConfigDP + ".serverConfig", serverConfig);
    
    if(dynlen(clientConfig) == 0 && dynlen(serverConfig) == 0) {
        dyn_int diNumber;
        dyn_string dsConfigFileDIPConfig;
        int iResult = _fwDIP_getAllDIPConfigFromConfigFile(diNumber, dsConfigFileDIPConfig);    // retrieve all the DIP Configs from config file
        
        iResult = dynContains(dsConfigFileDIPConfig, sFwDIPConfigDP);
        
        int res;
        
        if(iResult != 0) {   // _FwDIPConfig_FEName is in the config
            int iDIPManagerNum = diNumber[iResult]; // read the DIP manager number
            
            string sHost, sIP;
            int iPort;
            unGenericDpFunctions_getHostName(getSystemName(), sHost, sIP);
            unGenericDpFunctions_getpmonPort(getSystemName(), iPort);
            
            // try to remove manager
            res = fwInstallationManager_remove("WCCOAdip", "-num " + iDIPManagerNum, sHost, iPort);
            
            // remove entry in config file
            string configPath = getPath(CONFIG_REL_PATH, "config");
            res = paCfgDeleteValue(configPath, "dip_" + iDIPManagerNum, "DipConfig", sFwDIPConfigDP);
        }
        else {
            DebugN(sFwDIPConfigDP + " not found in config file");
        }
        
        // remove config
        res = dpDelete(sFwDIPConfigDP);
        DebugN("dpDelete result = " + res + " (0 = OK, -1 = error)");
    }
}

//------------------------------------------------------------------------------------------------------------------------

// DIP_getFrontEndArchiveDp
/**
Purpose: return the list of Dps that wil be archived during the set of the DIP front-end

Parameters:
		sCommand: string, input, the current entry in the import file
		sDp: string, input, the given dp: PLC or SystemAlarm  Dp
		sFEType: string, input, the front-end type
		dsArchiveDp: dyn_string, output, list of archivedDp

Usage: External function

PVSS manager usage: Ctrl, NG, NV

Constraints:
	. PVSS version: 2.12.1 
	. operating system: NT and W2000, but tested only under W2000.
	. distributed system: yes.
*/

DIP_getFrontEndArchiveDp(string sCommand, string sDp, string sFEType, dyn_string &dsArchiveDp)
{
	switch(sCommand) {
		case UN_PLC_COMMAND:
			dsArchiveDp = makeDynString();
			break;
		default:
			dsArchiveDp = makeDynString();
			break;
	}
}

// unConfigDIP_addSubscription
/**
Purpose: add a DIP subscription

Parameters:
		dpe: string, input, dp name
		frontEndName : string, input, FrontEnd name
		addressReference: string, input, address of the subscription
		exceptionInfo: dyn_string, output, for errors

Usage: External function

PVSS manager usage: Ctrl, NG, NV

Constraints:
	. PVSS version: 2.12.1 
	. operating system: NT and W2000, but tested only under W2000.
	. distributed system: yes.
*/

unConfigDIP_addSubscription(string dpe, string frontEndName, string addressReference, dyn_string& exceptionInfo)
{
    if(strpos(addressReference, "value=") == 0 || strpos(addressReference, "#VALUE!") == 0) {    // skip values specified as a value (value=) or not existing addresses (#VALUE!)
        DebugTN("unConfigDIP_addSubscription: Skipping subscription for: " + addressReference);
        return;
    }
    dyn_string configParameters = makeDynString();
    
    configParameters[fwPeriphAddress_TYPE] = fwPeriphAddress_TYPE_DIP;
    configParameters[fwPeriphAddress_DIP_CONFIG_DP] = "DIPConfig_"+ frontEndName;
    configParameters[fwPeriphAddress_ROOT_NAME] = addressReference;
        
    fwPeriphAddress_set(dpe, configParameters, exceptionInfo);
    if (dynlen(exceptionInfo) > 0) {
        DebugTN(exceptionInfo);
    }
}

// unConfigDIP_addPublication
/**
Purpose: add a DIP publication

Parameters:
		dpe: string, input, dp name
		frontEndName : string, input, FrontEnd name
		addressReference: string, input, address of the subscription
		exceptionInfo: dyn_string, output, for errors
        updateRate: int, input
        overwrite: bool, input

Usage: External function

PVSS manager usage: Ctrl, NG, NV

Constraints:
	. PVSS version: 2.12.1 
	. operating system: NT and W2000, but tested only under W2000.
	. distributed system: yes.
*/
unConfigDIP_addPublication(string dpe, string frontEndName, string addressReference, dyn_string& exceptionInfo, int updateRate=0, bool overwrite=false)
{
    if(strpos(addressReference, "value=") == 0 || strpos(addressReference, "#VALUE!") == 0) {    // skip values specified as a value (value=) or not existing addresses (#VALUE!)
        DebugTN("unConfigDIP_addPublication: Skipping publication for: " + addressReference);
        return;
    }
    
    string dipItem, dipTag, configDp = "DIPConfig_"+ frontEndName;

    _fwDIP_splitAddress(addressReference, dipItem, dipTag, exceptionInfo);

    // check if already published and unpublish if overwrite==true
    bool isConfigured;
    _fwDIP_checkIsItemPublished(dipItem, isConfigured, configDp, exceptionInfo);
    if(isConfigured && overwrite){
        fwDIP_unpublish(dipItem, exceptionInfo);
    }
    // TODO: check if can be published
    fwDIP_publishStructure(dipItem, makeDynString(dpe), makeDynString(dipTag), updateRate, configDp, exceptionInfo, overwrite);

    if (dynlen(exceptionInfo) > 0) {
        DebugTN(exceptionInfo);
    }
}

// DIP_Com_ExportConfig
/**
Purpose: function to export DIP FE configuration

Parameters:
		dsParam: dyn_string, input, parameters
		exceptionInfo: dyn_string, output, for errors
        
Usage: External function

PVSS manager usage: Ctrl, NG, NV

Constraints:
	. PVSS version: 2.12.1 
	. operating system: NT and W2000, but tested only under W2000.
	. distributed system: yes.
*/
DIP_Com_ExportConfig(dyn_string dsParam, dyn_string &exceptionInfo) {
    dyn_string dsDpParameters;
    dyn_string dsBoolArchive, dsAnalogArchive, dsEventArchive;
    string sPlcName, sPlcDrv;

    if (dynlen(dsParam) == UN_CONFIG_EXPORT_LENGTH) {
        sPlcName = dsParam[UN_CONFIG_EXPORT_PLCNAME];
        dpGet(c_unSystemIntegrity_UNDIP + sPlcName + ".configuration.archive_bool", dsBoolArchive,
              c_unSystemIntegrity_UNDIP + sPlcName + ".configuration.archive_analog", dsAnalogArchive,
              c_unSystemIntegrity_UNDIP + sPlcName + ".configuration.archive_event", dsEventArchive,
              c_unSystemIntegrity_UNDIP + sPlcName + ".communication.driver_num", sPlcDrv);

        //Delete
        dynClear(dsDpParameters);
        dynAppend(dsDpParameters, "#" + UN_DELETE_COMMAND);
        dynAppend(dsDpParameters, dsParam[UN_CONFIG_EXPORT_PLCNAME]);
        dynAppend(dsDpParameters, dsParam[UN_CONFIG_EXPORT_APPLICATION]);
        unExportDevice_writeDsStringToFile(dsDpParameters, UN_PARAMETER_DELIMITER);

        //PLCCONFIG
        dynClear(dsDpParameters);
        dynAppend(dsDpParameters, UN_PLC_COMMAND);
        dynAppend(dsDpParameters, "DIP");
        dynAppend(dsDpParameters, sPlcName);
        dynAppend(dsDpParameters, dsParam[UN_CONFIG_EXPORT_APPLICATION]);

        // Add DIP data
        // counter
        string counterAddr = DRV_DIP_convertToUnicosAddress(c_unSystemIntegrity_UNDIP + sPlcName + ".communication.counter", UNDIP_DPTYPE);
        dynAppend(dsDpParameters, counterAddr);
        
        string sDIPdns = "";
        string sOtherOptions = "";
        
        // retrieve DIP manager settings
        string sFwDIPConfigDP = "DIPConfig_" + sPlcName;
        
        dyn_int diNumber;
        dyn_string dsConfigFileDIPConfig;
        int iResult = _fwDIP_getAllDIPConfigFromConfigFile(diNumber, dsConfigFileDIPConfig);    // retrieve all the DIP Configs from config file
        if(iResult != 0) {
            fwException_raise(exceptionInfo,"ERROR","Cannot retrieve DIP config info from config file.","");
        } else {
            iResult = dynContains(dsConfigFileDIPConfig, sFwDIPConfigDP); // check if _FwDIPConfig_FEName is already in use inside config
        
            if(iResult == 0) {
                fwException_raise(exceptionInfo,"ERROR","No dip manager for config " + sFwDIPConfigDP + "!","");
            } else {
                int iDIPManagerNum = diNumber[iResult]; // read the DIP manager number
            
                string sHost, sIP;
                int iPort;
                unGenericDpFunctions_getHostName(getSystemName(), sHost, sIP);
                unGenericDpFunctions_getpmonPort(getSystemName(), iPort);
        
                dyn_mixed managerInfo;
                fwInstallationManager_getProperties("WCCOAdip", "-num " + iDIPManagerNum, managerInfo, sHost, iPort);
                string sProperties = managerInfo[FW_INSTALLATION_MANAGER_OPTIONS];
                dyn_string dsProperties = strsplit(sProperties, "-");
                
                dyn_string dsOtherOptions;
                
                for(int i=1; i<=dynlen(dsProperties); i++){
                    if(strpos(strtolower(dsProperties[i]), "num") == 0) continue;   // skip -num parameter
                    else if(strpos(strtolower(dsProperties[i]), "dns") == 0) {      // special for dns parameter
                        sDIPdns = substr(dsProperties[i], 4, strlen(dsProperties[i]) - 4);
                        // remove escaped quotes
                        strreplace(sDIPdns, "\\", "");
                        strreplace(sDIPdns, "\"", "");
                    } else if (dsProperties[i] != ""){
                        dynAppend(dsOtherOptions, "-" + dsProperties[i]);
                    }
                }
                sOtherOptions = strjoin(dsOtherOptions, " ");
            }
        }
        
        // dip DNSes
        dynAppend(dsDpParameters, sDIPdns);
        // additional parameters
        dynAppend(dsDpParameters, sOtherOptions);
        
        // publish name
        string publisherName;
        dpGet(sFwDIPConfigDP + ".publishName", publisherName);
        dynAppend(dsDpParameters, publisherName);
        
        unExportDevice_writeDsStringToFile(dsDpParameters, UN_PARAMETER_DELIMITER);
    } else {
        fwException_raise(exceptionInfo, "ERROR", "SNMP_Com_ExportConfig(): Wrong Parameters", "");
    }
}

// DRV_DIP_convertToUnicosAddress
/**
Purpose: function to export DIP FE configuration

Parameters:
		sDeviceDpeName: string, input, dpe name for which address should be formatted
		sDeviceType: string, input, device type name
        
Usage: External function

PVSS manager usage: Ctrl, NG, NV

Constraints:
	. PVSS version: 2.12.1 
	. operating system: NT and W2000, but tested only under W2000.
	. distributed system: yes.
*/
string DRV_DIP_convertToUnicosAddress(string sDeviceDpeName, string sDeviceType)
{
    int configType;
    mixed addr_ref;
    dpGet(sDeviceDpeName + ":_address.._type", configType,
          sDeviceDpeName + ":_address.._reference" , addr_ref);
    if (configType == DPCONFIG_NONE) { // if addr is not defined - return an empty string
        return "";
    }

    return addr_ref;
}

//------------------------------------------------------------------------------------------------------------------------
#uses "unDIP.ctl"
#uses "fwDIP/fwDIPImport.ctl"
//@}
