/**
@file 
@date        11/03/2015
@author      Josef Hofer 
@copyright   CERN (EN-ICE-SCD) 

@brief Library contains functions used by the unLHCLaser component.

Usage: Public

PVSS manager usage: CTRL, UI

Constraints:
- PVSS version: 3.11
- operating system: Linux.
*/

// Input commands
const int LASER_MANAGER_COMMAND_IN_UNKNOWN = 0;
const int LASER_MANAGER_COMMAND_IN_START   = 10;
const int LASER_MANAGER_COMMAND_IN_ABORT   = 99;
const int LASER_MANAGER_COMMAND_IN_STOP    = 100;

// Manager states
const int LASER_MANAGER_STATE_UNKNOWN   = 1000;
const int LASER_MANAGER_STATE_INIT  	= 1005; 
const int LASER_MANAGER_STATE_ACTIVE 	= 1010;	
const int LASER_MANAGER_STATE_PASSIVE	= 1040;	
const int LASER_MANAGER_STATE_STOPPED 	= 1050;	
const int LASER_MANAGER_STATE_SUSPENDED = 1080;	
const int LASER_MANAGER_STATE_FINISHED 	= 1090;

// Manager infos, warnings and errors
const int LASER_MANAGER_MSG_INFO_BASE   = 10000;
const int LASER_MANAGER_MSG_WARNING_BASE= 10100;
const int LASER_MANAGER_MSG_ERROR_BASE  = 10200;

const int LASER_CONFIGPATH=1;
const int LASER_TCP_REMOTEHOST=2;
const int LASER_TCP_DATAPATH=3;
const int LASER_SHOWSTATISTICS=4;
const int LASER_TCP_MAXPACKETSIZE=5;
const int LASER_BACKUP_PERIOD=6;
const int LASER_LOG_LEVEL=7;
const int LASER_NUM_PERIOD=8;
const int LASER_T_MODE=9;

//State Manager
const int Laser_Command=1;
const int Laser_StateOut=2;
const int Laser_Message=3;
const int Laser_PVSS=4;

//Catalogue name
const string g_sCat_Laser="unLaser";

//user allowed to modify LASER if Unicos Package not installed
const string g_Laser_userAdmin="root"; 

// default settings
private const string DEFAULT_LASER_BROKER_PORT      = "2580";
private const string DEFAULT_LASER_CONFIG_FILE      = "data/LaserData.xml";
private const string DEFAULT_LASER_BROKER_LIST      = "LASER-TEST:2580";
private const string DEFAULT_LASER_TOPIC_BASE       = "EN_ICE_TEST";
private const string DEFAULT_LASER_STATISTIC_MODE   = "200";
private const string DEFAULT_LASER_T_MODE_SETTING   = "FALSE";
private const string DEFAULT_LASER_MAX_PACKET_SIZE  = "1000";
private const string DEFAULT_LASER_BACKUP_PERIOD    = "600";
private const string DEFAULT_LASER_LOG_LEVEL        = "6";
private const string DEFAULT_LASER_HEARTBEAT_PERIOD = "10";


/** Get laser manager default settings
@param  dsExInfo     output, error information
@return dsDefaultSettings   the settings
*/
public dyn_string unLaser_getManagerDefaultSettings(dyn_string &dsExInfo)
{    
    dyn_string dsDefaultSettings;
   	string sSystemName = getSystemName();
	
   	if (dpExists(sSystemName + "Laser_Info.defaultSettings")) {
		      int iRes=dpGet(sSystemName + "Laser_Info.defaultSettings.Settings", dsDefaultSettings);
        if (iRes < 0) {
			         fwException_raise(dsExInfo, "ERROR", __FUNCTION__ + ": " + getCatStr(g_sCat_Laser,"DPSETGETFAILED"),"");	
        } else {            
            dyn_string dsHardcodedDefaultSettings = unLaser_getHardcodedDefaultSettings();                        
            if(dynlen(dsDefaultSettings) != dynlen(dsHardcodedDefaultSettings)) {
                dsDefaultSettings = dsHardcodedDefaultSettings;
                iRes=dpSetWait(sSystemName + "Laser_Info.defaultSettings.Settings", dsDefaultSettings);                
            }         
        }       		
	   } else {        		
		      int iRes=dpCreate("Laser_Info","_Un_Laser_Config");
		      if (iRes < 0) {
			         fwException_raise(dsExInfo, "ERROR", __FUNCTION__ + ": " + getCatStr(g_sCat_Laser,"DPCREATEFAILURE"),"");	
		      } else {	
            dsDefaultSettings = unLaser_getHardcodedDefaultSettings();																																																																				
			         iRes=dpSetWait(sSystemName + "Laser_Info.defaultSettings.Settings", dsDefaultSettings);
        }		
    }
    
    return dsDefaultSettings;
}


private dyn_string unLaser_getHardcodedDefaultSettings()
{
    dyn_string dsSettings;
   	dsSettings[LASER_CONFIGPATH] = DEFAULT_LASER_CONFIG_FILE;
   	dsSettings[LASER_TCP_REMOTEHOST] = DEFAULT_LASER_BROKER_LIST;
   	dsSettings[LASER_TCP_DATAPATH] = DEFAULT_LASER_TOPIC_BASE;
   	dsSettings[LASER_SHOWSTATISTICS] = DEFAULT_LASER_STATISTIC_MODE;
    dsSettings[LASER_T_MODE] = DEFAULT_LASER_T_MODE_SETTING;
   	dsSettings[LASER_TCP_MAXPACKETSIZE] = DEFAULT_LASER_MAX_PACKET_SIZE;
   	dsSettings[LASER_BACKUP_PERIOD] = DEFAULT_LASER_BACKUP_PERIOD;
   	dsSettings[LASER_LOG_LEVEL] = DEFAULT_LASER_LOG_LEVEL;
   	dsSettings[LASER_NUM_PERIOD] = DEFAULT_LASER_HEARTBEAT_PERIOD;
    return dsSettings;        
}


/** Set laser manager settings to default settings
@param sDP          input, Laser Manager DP 
@param dsExInfo     output, error information
*/
public void unLaser_setManagerDefaultSettings(string sDP, dyn_string &dsExInfo)
{
    dyn_string dsSettings = unLaser_getManagerDefaultSettings(dsExInfo);

    if (dynlen(dsExInfo)==0) {
	       if (dpExists(sDP + ".Settings")) {
          		int iRes=dpSetWait(
                sDP + ".Settings.man_ConfigPath", dsSettings[LASER_CONFIGPATH],
            				sDP + ".Settings.man_ShowStatistics", dsSettings[LASER_SHOWSTATISTICS],
            				sDP + ".Settings.man_BackupPeriod", dsSettings[LASER_BACKUP_PERIOD],
            				sDP + ".Settings.tcp_LogLevel", dsSettings[LASER_LOG_LEVEL],
            				sDP + ".Settings.tcp_BrokerName", dsSettings[LASER_TCP_REMOTEHOST],
            				sDP + ".Settings.tcp_TopicBase", dsSettings[LASER_TCP_DATAPATH],
            				sDP + ".Settings.T_mode", dsSettings[LASER_T_MODE],        
            				sDP + ".Settings.tcp_MaxPacketSize", dsSettings[LASER_TCP_MAXPACKETSIZE],
            				sDP + ".Settings.man_NumPeriod", dsSettings[LASER_NUM_PERIOD],                
            				sDP + ".State.st_CommandIn", LASER_MANAGER_COMMAND_IN_STOP,			
            				sDP + ".State.st_StateOut", LASER_MANAGER_STATE_UNKNOWN,
            				sDP + ".State.ev_MessageOut", LASER_MANAGER_MSG_INFO_BASE);		//Unknow message
    		  } else {
		          fwException_raise(dsExInfo, "ERROR", __FUNCTION__ + ": " + getCatStr(g_sCat_Laser,"DPDOESNOTEXIST"),"");
        }
    }
}


/** Get laser manager settings
@param  sDP         input, Laser Manager DP 
@param  dsExInfo    output, error information
@return dsSettings  the requested settings       
*/
public dyn_string unLaser_getManagerSettings(string sDP, dyn_string &dsExInfo)
{
    dyn_string dsSettings; 
    
    if (dpExists(sDP + ".Settings")) {
	       int iRes=dpGet(
            sDP + ".Settings.man_ConfigPath", dsSettings[LASER_CONFIGPATH],
        				sDP + ".Settings.man_ShowStatistics", dsSettings[LASER_SHOWSTATISTICS],
        				sDP + ".Settings.man_BackupPeriod", dsSettings[LASER_BACKUP_PERIOD],
        				sDP + ".Settings.tcp_LogLevel", dsSettings[LASER_LOG_LEVEL],
        				sDP + ".Settings.tcp_BrokerName", dsSettings[LASER_TCP_REMOTEHOST],
        				sDP + ".Settings.tcp_TopicBase", dsSettings[LASER_TCP_DATAPATH],
        				sDP + ".Settings.tcp_MaxPacketSize", dsSettings[LASER_TCP_MAXPACKETSIZE],
        				sDP + ".Settings.man_NumPeriod", dsSettings[LASER_NUM_PERIOD],
            sDP + ".Settings.T_mode", dsSettings[LASER_T_MODE]);

        if (iRes<0) {
		          fwException_raise(dsExInfo, "ERROR", __FUNCTION__ + ": " + getCatStr(g_sCat_Laser,"DPGETERROR"),"");
        }		
    } else {
	       fwException_raise(dsExInfo, "ERROR", __FUNCTION__ + ": " + getCatStr(g_sCat_Laser,"DPDOESNOTEXIST"),"");
    }	
    
    return dsSettings;
}

/** Check the specified server list and add default port number if it not exists.
@param  serverList       broker list to fix
@return fixedServerList  fixed broker list       
*/
public string unLaser_fixServerList(string serverList)
{
  string fixedServerList = "";
  dyn_string servers = strsplit(serverList, ",");
  
  for(int i=1; i<=dynlen(servers); i++)
  {
    dyn_string splitServerPort = strsplit(servers[i], ":");
    //Check if port is defined, if not add the default port
    if(dynlen(splitServerPort) < 2) {
      servers[i] = servers[i] + ":" + DEFAULT_LASER_BROKER_PORT;
    }
    //Append server to the list
    fixedServerList += servers[i];    
    
    //If we are not in the last Server of the list , add comma
    if(i != dynlen(servers)) {
       fixedServerList += ",";
    }
  }

  return fixedServerList;
}


//Function to  hide the full path of the config file and to convert it to "LastFolder\Filename" format.
public string unLaser_hideFilePath(string filename)
{
  string sPath = getPath(DATA_REL_PATH);
		
  if ((filename != "") && (strpos (filename, sPath)==0)) {
    int index=strpos (filename, "data");		
    sPath=substr(filename, 0, index);
    sPath=substr(filename, index, strlen(filename)-strlen(sPath));
  }
  return sPath;
}


// Extract manager number from manager name
public int unLaser_getManagerNumberFromName(string sManagerName)
{
    int iManagerNumber;
	   string sNum = substr(sManagerName, strpos(sManagerName,c_Laser_Pattern) + strlen(c_Laser_Pattern));
	   sscanf(sNum,"%d",iManagerNumber);    
    return iManagerNumber;    
}
