//------------------------------------
// Library uses
//------------------------------------

//------------------------------------
// Constant declaration
//------------------------------------
const int c_unSystemIntegrity_IEC104_defaultCheckingDelay = 10;
const int c_unSystemIntegrity_IEC104_numberOfParameters   = 1;

const string UN_SYSTEM_INTEGRITY_IEC104       = "IEC104";
const string UN_SYSTEM_INTEGRITY_IEC104_check = "_unSystemIntegrity_IEC104_Check";
const string UN_IEC104_DPTYPE                 =  "IEC104";
const string UN_IEC104_COMPONENT_NAME         = "unFE_IEC104";

const int iSYSTEM_INTEGRITY_IEC104_ERROR_DRIVER_TYPE                  = 20;
const int iSYSTEM_INTEGRITY_IEC104_ERROR_MANAGER_STOPPED              = 21;
const int iSYSTEM_INTEGRITY_IEC104_ERROR_CONNECTION_TO_HOST           = 22;
const int iSYSTEM_INTEGRITY_IEC104_ERROR_CONNECTION_NOT_READY         = 23;
const int iSYSTEM_INTEGRITY_IEC104_ERROR_CONNECTION_DEVICE_NOT_READY  = 24;


//------------------------------------
// Global declaration
//------------------------------------
global dyn_string g_dsUnSystemIntegrity_IEC104_FrontEndList;      // List of IEC104 FrontEnd DP that are monitored
global dyn_int    g_diUnSystemIntegrity_IEC104_FrontEndThreadId;  // List of IEC104 FrontEnd thread Ids, that are monitoring
global int        g_i_unSystemIntegrity_IEC104_CheckingDelay;      // Current delay between two checks

//@{





// IEC104_systemIntegrityInfo
/** Return the list of systemAlarm pattern.

@return list of systemAlarm pattern

*/
dyn_string IEC104_systemIntegrityInfo()
{

  return makeDynString(DS_pattern);

}// IEC104_systemIntegrityInfo()





// unSystemIntegrity_IEC104_Initialize
/**
Purpose: Get the list of IEC104 front ends checked by the systemIntegrity. Only the ones that are enabled.

@param dsResult: dyn_string, output, the list of enabled IEC104 system integrities that are being checked

*/
void unSystemIntegrity_IEC104_Initialize(dyn_string &dsResult)
{
  bool bEnabled;
  int iLen, iLoop;
  string sDpToCheck;
  dyn_string dsList;
  dyn_errClass deError;


  // Update default configuration data
  sDpToCheck = UN_SYSTEM_INTEGRITY_IEC104 + UN_SYSTEMINTEGRITY_EXTENSION + ".config.data";
  if( dpGet(sDpToCheck, dsList) != 0 )
  {
    DebugTN("unSystemIntegrity_IEC104_Initialize() -> ERROR -> Error getting the configuration data for the IEC104 FrontEnd from DP: " + sDpToCheck);
    return;
  }
  else
  {
    deError = getLastError();
    if( dynlen(deError) > 0 )
    {
      DebugTN("unSystemIntegrity_IEC104_Initialize() -> ERROR -> Error getting the configuration data for the IEC104 FrontEnd from DP: " + sDpToCheck + ", due to: " + getErrorText(deError));
      return;
    }
  }
  _unSystemIntegrity_IEC104_DataCallback("", dsList);


  // Get the list of enabled and running IEC104 system integrities
  dsList  = dpNames(c_unSystemAlarm_dpPattern + DS_pattern + "*", c_unSystemAlarm_dpType);
  deError = getLastError();
  if( dynlen(deError) > 0 )
  {
    DebugTN("unSystemIntegrity_IEC104_Initialize() -> ERROR -> Error getting the list of IEC104 system integrity DPs, due to: " + getErrorText(deError));
    return;
  }

  iLen = dynlen(dsList);
  for( iLoop = 1 ; iLoop <= iLen ; iLoop++ )
  {
    sDpToCheck = dpSubStr(dsList[iLoop], DPSUB_DP);
    sDpToCheck = substr(sDpToCheck, strlen(c_unSystemAlarm_dpPattern + DS_pattern), strlen(sDpToCheck));

    if( (dpExists(sDpToCheck))                       &&
        (dpTypeName(sDpToCheck) == UN_IEC104_DPTYPE)   )
    {
      if( dpGet(dsList[iLoop] + ".enabled", bEnabled) != 0 )
      {
        DebugTN("unSystemIntegrity_IEC104_Initialize() -> ERROR -> Error getting IEC104 system integrity enable status: " + dsList[iLoop]);
        return;
      }
      else
      {
        deError = getLastError();
        if( dynlen(deError) > 0 )
        {
          DebugTN("unSystemIntegrity_IEC104_Initialize() -> ERROR -> Error getting IEC104 system integrity enable status: " + dsList[iLoop] + " due to: " + getErrorText(deError));
          return;
        }

        if( bEnabled )
        {
          dynAppend(dsResult, sDpToCheck);
        }
      }
    }
  }

}//  unSystemIntegrity_IEC104_Initialize()





// _unSystemIntegrity_IEC104_DataCallback
/**
Purpose: handle any systemIntegrity parameter change

@param sDpe: string, input, dpe
@param dsConfigData: dyn_string, input, the config data

*/
void _unSystemIntegrity_IEC104_DataCallback(string sDpe, dyn_string dsConfigData)
{
  dyn_string dsTemp;


  dsTemp = dsConfigData;
  // Fill dsTemp with enough parameters if the Data callback is incomplete
  while( dynlen(dsTemp) < c_unSystemIntegrity_IEC104_numberOfParameters )
  {
    dynAppend(dsTemp, "0");
  }

  g_i_unSystemIntegrity_IEC104_CheckingDelay = (int) dsTemp[1];
  if( g_i_unSystemIntegrity_IEC104_CheckingDelay <= 0 )
  {
    g_i_unSystemIntegrity_IEC104_CheckingDelay = c_unSystemIntegrity_IEC104_defaultCheckingDelay;
  }

}//  _unSystemIntegrity_IEC104_DataCallback()





// unSystemIntegrity_IEC104_HandleCommand
/**
Purpose: handle any systemIntegrity command.

@param sDpeCommand: string, input, dpe
@param iCommand: int, input, the systemIntegrity command
@param sDpeParameters: string, input, dpe
@param dsParameters: dyn_string, input, the list of parameters of the systemIntegrity command

*/
void unSystemIntegrity_IEC104_HandleCommand(string sDpeCommand, int iCommand, string sDpeParameters, dyn_string dsParameters)
{
  int iLoop, iLen;
  string sResult;
  dyn_string dsResult, exceptionInfo;
  dyn_errClass deError;


  iLen = dynlen(dsParameters);

  switch( iCommand )
  {
    case UN_SYSTEMINTEGRITY_ADD: // Add the _UnSystemAlarm DP
      for( iLoop = 1 ; iLoop <= iLen ; iLoop++ )
      {
        _unSystemIntegrity_IEC104_Checking(dsParameters[iLoop], TRUE, TRUE, exceptionInfo);
        if( dynlen(exceptionInfo) > 0 )
        {
          _unSystemIntegrity_IEC104_sendException("unSystemIntegrity_IEC104_HandleCommand",
                                                 "ERROR",
                                                 "Error checking parameter[" + iLoop + "] parameter: " + dsParameters[iLoop] + " failled due to: " + exceptionInfo[2]);
          return;
        }
      }
      break;


    case UN_SYSTEMINTEGRITY_DELETE: // Delete the _UnSystemAlarm DP
      for( iLoop = 1 ; iLoop <= iLen ; iLoop++ )
      {
        _unSystemIntegrity_IEC104_Checking(dsParameters[iLoop], FALSE, FALSE, exceptionInfo);
        if( dynlen(exceptionInfo) > 0 )
        {
          _unSystemIntegrity_IEC104_sendException("unSystemIntegrity_IEC104_HandleCommand",
                                                 "ERROR",
                                                 "Error disabling _UnSystemAlarm: " + dsParameters[iLoop] + ", failled due to: " + exceptionInfo[2]);
          return;
        }

        // Remove the DP from the alert list of the Application DP if it is in
        _unSystemIntegrity_modifyApplicationAlertList(c_unSystemAlarm_dpPattern + DS_pattern + dsParameters[iLoop], FALSE, exceptionInfo);
        if( dynlen(exceptionInfo) > 0 )
        {
          _unSystemIntegrity_IEC104_sendException("unSystemIntegrity_IEC104_HandleCommand",
                                                 "ERROR",
                                                 "Error removing system alarm: " + dsParameters[iLoop] + ", from the Application alarm list, due to: " + exceptionInfo[2]);
          return;
        }

        // Delete the _UnSystemAlarm dps
        if( dpExists(c_unSystemAlarm_dpPattern + DS_pattern + dsParameters[iLoop]) )
        {
          if( dpDelete(c_unSystemAlarm_dpPattern + DS_pattern + dsParameters[iLoop]) != 0 )
          {
            _unSystemIntegrity_IEC104_sendException("unSystemIntegrity_IEC104_HandleCommand",
                                                   "ERROR",
                                                   "Error removing sysmtem alarm DP: system alarm: " + c_unSystemAlarm_dpPattern + DS_pattern + dsParameters[iLoop]);
            return;
          }
          else
          {
            deError = getLastError();
            if( dynlen(deError) > 0 )
            {
              _unSystemIntegrity_IEC104_sendException("unSystemIntegrity_IEC104_HandleCommand",
                                                     "ERROR",
                                                     "Error removing sysmtem alarm DP: system alarm: " + c_unSystemAlarm_dpPattern + DS_pattern + dsParameters[iLoop] + ", due to: " + getErrorText(deError));
              return;
            }
          }
        }
      }
      break;


    case UN_SYSTEMINTEGRITY_ENABLE: // Enable the _UnSystemAlarm DP
      for( iLoop = 1 ; iLoop <= iLen ; iLoop++ )
      {
        if( dpExists(c_unSystemAlarm_dpPattern + DS_pattern + dsParameters[iLoop]) )
        {
          _unSystemIntegrity_IEC104_Checking(dsParameters[iLoop], FALSE, TRUE, exceptionInfo);
          if( dynlen(exceptionInfo) > 0 )
          {
            _unSystemIntegrity_IEC104_sendException("unSystemIntegrity_IEC104_HandleCommand",
                                                   "ERROR",
                                                   "Error enabling _UnSystemAlarm: " + dsParameters[iLoop] + ", failled due to: " + exceptionInfo[2]);
            return;
          }
        }
      }
      break;


    case UN_SYSTEMINTEGRITY_DISABLE: // Disable the _UnSystemAlarm DP
      for( iLoop = 1 ; iLoop <= iLen ; iLoop++ )
      {
        if( dpExists(c_unSystemAlarm_dpPattern + DS_pattern + dsParameters[iLoop]) )
        {
          _unSystemIntegrity_IEC104_Checking(dsParameters[iLoop], FALSE, FALSE, exceptionInfo);
          if( dynlen(exceptionInfo) > 0 )
          {
            _unSystemIntegrity_IEC104_sendException("unSystemIntegrity_IEC104_HandleCommand",
                                                   "ERROR",
                                                   "Error disabling _UnSystemAlarm: " + dsParameters[iLoop] + ", failled due to: " + exceptionInfo[2]);
            return;
          }
        }
      }
      break;


    case UN_SYSTEMINTEGRITY_DIAGNOSTIC: // Give the list of _UnSystemAlarm DP and the state
      if( dpSetWait(UN_SYSTEM_INTEGRITY_IEC104 + UN_SYSTEMINTEGRITY_EXTENSION + ".diagnostic.commandResult", iCommand,
                    UN_SYSTEM_INTEGRITY_IEC104 + UN_SYSTEMINTEGRITY_EXTENSION + ".diagnostic.result",        g_dsUnSystemIntegrity_IEC104_FrontEndList) != 0 )
      {
        _unSystemIntegrity_IEC104_sendException("unSystemIntegrity_IEC104_HandleCommand",
                                               "ERROR",
                                               "Error setting diagnostic command: " + iCommand + " with parameters: " + g_dsUnSystemIntegrity_IEC104_FrontEndList);
        return;
      }
      else
      {
        deError = getLastError();
        if( dynlen(deError) > 0 )
        {
          _unSystemIntegrity_IEC104_sendException("unSystemIntegrity_IEC104_HandleCommand",
                                                 "ERROR",
                                                 "Error setting diagnostic command: " + iCommand + " with parameters: " + g_dsUnSystemIntegrity_IEC104_FrontEndList + ", due to: " + getErrorText(deError));
          return;
        }
      }
      break;


    case UN_SYSTEMINTEGRITY_EXTENDED_DIAGNOSTIC: // Extended diagnostic
      _unSystemIntegrity_IEC104_ExtendedDiagnostic();
      break;


    case UN_SYSTEMINTEGRITY_DEBUG:
      _unSystemIntegrity_IEC104_Debug(dsParameters);
      break;


    case UN_SYSTEMINTEGRITY_VERSION:
      _unSystemIntegrity_IEC104_Version();
      break;

    default:
      _unSystemIntegrity_IEC104_sendException("unSystemIntegrity_IEC104_HandleCommand",
                                             "ERROR",
                                             "Error, unknown command: " + iCommand);
      return;
      break;
  }

}// unSystemIntegrity_IEC104_HandleCommand()





// _unSystemIntegrity_IEC104_Version
/**
Purpose: Returns the component version via system integrity command answer

*/
void _unSystemIntegrity_IEC104_Version()
{
  string sMessage;


  sMessage = unGenericDpFunctions_getComponentsVersion(makeDynString("unFE_IEC104"));

  dpSet(UN_SYSTEM_INTEGRITY_IEC104 + UN_SYSTEMINTEGRITY_EXTENSION + ".diagnostic.commandResult", UN_SYSTEMINTEGRITY_VERSION,
        UN_SYSTEM_INTEGRITY_IEC104 + UN_SYSTEMINTEGRITY_EXTENSION + ".diagnostic.result",        makeDynString(getCurrentTime(), sMessage));

}//  _unSystemIntegrity_IEC104_Version()





// _unSystemIntegrity_IEC104_ExtendedDiagnostic
/**
Purpose: Returns the list of interesting variables and to debug the system integrity via system integrity command answer

*/
void _unSystemIntegrity_IEC104_ExtendedDiagnostic()
{
  int iLoop, iLen;
  string sResult;
  dyn_string dsResult;
  dyn_errClass deError;


  dynAppend(dsResult, getCurrentTime());

  dynAppend(dsResult, "g_dsUnSystemIntegrity_IEC104_FrontEndList");
  iLen = dynlen(g_dsUnSystemIntegrity_IEC104_FrontEndList);
  for( iLoop = 1 ; iLoop <= iLen ; iLoop++ )
  {
    sResult = g_dsUnSystemIntegrity_IEC104_FrontEndList[iLoop];
    dynAppend(dsResult, sResult);
  }

  dynAppend(dsResult, "g_diUnSystemIntegrity_IEC104_FrontEndThreadId");
  iLen = dynlen(g_diUnSystemIntegrity_IEC104_FrontEndThreadId);
  for( iLoop = 1 ; iLoop <= iLen ; iLoop++ )
  {
    sResult = g_diUnSystemIntegrity_IEC104_FrontEndThreadId[iLoop];
    dynAppend(dsResult, sResult);
  }

  dynAppend(dsResult, "g_i_unSystemIntegrity_IEC104_CheckingDelay");
  dynAppend(dsResult, g_i_unSystemIntegrity_IEC104_CheckingDelay);

  dynAppend(dsResult, "g_b32DebugLevel");
  dynAppend(dsResult, g_b32DebugLevel);

  dynAppend(dsResult, "g_sDebugFilter");
  dynAppend(dsResult, g_sDebugFilter);

  if( dpSet(UN_SYSTEM_INTEGRITY_IEC104 + UN_SYSTEMINTEGRITY_EXTENSION + ".diagnostic.commandResult", UN_SYSTEMINTEGRITY_EXTENDED_DIAGNOSTIC,
            UN_SYSTEM_INTEGRITY_IEC104 + UN_SYSTEMINTEGRITY_EXTENSION + ".diagnostic.result",        dsResult) != 0 )
  {
    DebugTN("_unSystemIntegrity_IEC104_ExtendedDiagnostic() -> ERROR -> Error publishing answer to command: " + dsResult);
    return;
  }
  else
  {
    deError = getLastError();
    if( dynlen(deError) > 0 )
    {
      DebugTN("_unSystemIntegrity_IEC104_ExtendedDiagnostic() -> ERROR -> Error publishing answer to command: " + dsResult + ", due to: " + getErrorText(deError));
      return;
    }
  }

}//  _unSystemIntegrity_IEC104_ExtendedDiagnostic()





// _unSystemIntegrity_IEC104_Debug
/**
Purpose: Modifies the debug filters to log any runtime value

@param dsParameters: dyn_string, input, list of parameters passed by the UI panel:
                                        dsParameters[1] = Debug level (bit32)
                                        dsParameters[2] = Debug filter (string)

*/
void _unSystemIntegrity_IEC104_Debug(dyn_string dsParameters)
{
  dyn_errClass deError;


  if( dynlen(dsParameters) >= 2 )
  {
    g_b32DebugLevel = (bit32)dsParameters[1];
    g_sDebugFilter  =        dsParameters[2];
  }
  else
  {
    g_b32DebugLevel = (bit32)0;
    g_sDebugFilter  = "";
  }

  if( dpSet(UN_SYSTEM_INTEGRITY_IEC104 + UN_SYSTEMINTEGRITY_EXTENSION + ".diagnostic.commandResult", UN_SYSTEMINTEGRITY_DEBUG,
            UN_SYSTEM_INTEGRITY_IEC104 + UN_SYSTEMINTEGRITY_EXTENSION + ".diagnostic.result",        makeDynString(g_b32DebugLevel, g_sDebugFilter)) != 0 )
  {
    DebugTN("_unSystemIntegrity_IEC104_Debug() -> ERROR -> Error setting the debug filtering");
    return;
  }
  else
  {
    deError = getLastError();
    if( dynlen(deError) > 0 )
    {
      DebugTN("_unSystemIntegrity_IEC104_Debug() -> ERROR -> Error setting the debug filtering, due to: " + getErrorText(deError));
      return;
    }
  }
}// _unSystemIntegrity_IEC104_Debug()




// _unSystemIntegrity_IEC104_sendException
/**
Purpose: Shows error in the log and sent the exception into event log

@param sSender: string, input, Function sender
@param sErrorType: string, input, Error type: FATAL, ERROR, WARNING, INFO
@param sErrorMessage: string, input, Error message
@param sErrorCode: string, input Error code, by default empty

@return Nothing
*/
void _unSystemIntegrity_IEC104_sendException(string sSender, string sErrorType, string sErrorMessage, string sErrorCode = "")
{
  dyn_string exceptionInfo;


  fwException_raise(exceptionInfo, sErrorType, sErrorMessage, sErrorCode);
  DebugTN(sSender + " -> " + sErrorType + " -> " + sErrorMessage);
  unMessageText_sendException("*", "*", sSender, "user", "*", exceptionInfo);

}// _unSystemIntegrity_IEC104_sendException()





// _unSystemIntegrity_IEC104_Checking
/**
Purpose: This function register/de-register the callback funtion of the IEC104 manager.
This function can also create the _unSystemAlarm_IEC104 DP with the alarm config but it cannot delete it.

  @param sDp: string, input, data point name
  @param bCreate: bool, input, true create the dp and the alarm config if it does not exist, enable it
  @param bRegister: bool, input, true do a dpConnect, false do a dpDisconnect
  @param exceptionInfo: dyn_string, output, exception are returned here

@Return Nothing
*/
void _unSystemIntegrity_IEC104_Checking(string sDp, bool bCreate, bool bRegister, dyn_string &exceptionInfo)
{
  int iRes, iThreadId, iPos, iDriverNumber;
  string sDpToCheck, sFrontEndHostname, sDescription, sFrontEndDp, sFrontEndName, sFrontEndConnection, sFrontEndDriver;
  dyn_errClass deError;


  // Get the front end: DP, name, connection and driver
  sFrontEndDp   = sDp;
  sFrontEndName = sDp;

  // Remove the IEC104_ prefix
  if( substr(sFrontEndName, 0, 7 ) == "IEC104_" )
  {
    sFrontEndName = substr(sFrontEndName, 7);
  }

  sFrontEndConnection = "_IEC_" + sFrontEndName;

  // Remove the system name
  iRes = strpos(sDp, ":");
  if( iRes >= 0 )
  {
    sDp = substr(sDp, iRes + 1, strlen(sDp));
  }
  else
  {
    sDp = sDp;
  }

  sDpToCheck = sDp;
  sDp        = c_unSystemAlarm_dpPattern + DS_pattern + sDp;


  // In case the Dp is not of type IEC104 do nothing (should not happen, just for security)
  if( dpTypeName(sDpToCheck) != UN_IEC104_DPTYPE)
  {
    _unSystemIntegrity_IEC104_sendException("_unSystemIntegrity_IEC104_Checking", "ERROR", "Error DP to check: " + sDpToCheck + " is not of IEC104 type");
    return;
  }


  if( bCreate)
  {
    // Get the FrontEnd hostname and the driver number
    sFrontEndHostname = unGenericDpFunctions_getAlias(sDpToCheck);

    // Create the sDp and its alarm config if it is not existing
    if( dpExists(sDp) == FALSE )
    {
      sDescription = getCatStr("unSystemIntegrity", "PLC_DS_DESCRIPTION") + sFrontEndHostname + " -> DS driver " + iDriverNumber;
      unSystemIntegrity_createSystemAlarm(sDpToCheck, DS_pattern, sDescription, exceptionInfo);
      if( dynlen(exceptionInfo) > 0 )
      {
        _unSystemIntegrity_IEC104_sendException("_unSystemIntegrity_IEC104_Checking", "ERROR", "Error creating system integrity alarm for DP: " + sDpToCheck + ", due to: " + exceptionInfo[2]);
        return;
      }
    }
  }


  if( bRegister )
  {
    // Connect to the callback function if not already done
    iPos = _unSystemIntegrity_isInList(g_dsUnSystemIntegrity_IEC104_FrontEndList, sDp);
    if( iPos <= 0 )
    {
      // Add it in the list, because one of the callback function needs it.
      iPos = _unSystemIntegrity_setList(g_dsUnSystemIntegrity_IEC104_FrontEndList, sDp, TRUE);

      // Start the thread for checking the status
        // Get driver number, Connection DP and driver DP
      if( dpExists(sFrontEndDp) == FALSE )
      {
        _unSystemIntegrity_IEC104_sendException("_unSystemIntegrity_IEC104_Checking", "ERROR", "Error front end DP: "+ sFrontEndDp + " doesn't exist");
        return;
      }


      if( dpExists(sFrontEndConnection) == FALSE )
      {
        _unSystemIntegrity_IEC104_sendException("_unSystemIntegrity_IEC104_Checking", "ERROR", "Error front end IEC connection: " + sFrontEndConnection + " doesn't exist");
        return;
      }


      if( dpGet(sFrontEndDp + ".communication.driver_num", iDriverNumber) != 0 )
      {
        _unSystemIntegrity_IEC104_sendException("_unSystemIntegrity_IEC104_Checking", "ERROR", "Error getting the driver number, IEC connection and IEC driver " + sFrontEndDp + ".communication.driver_num");
        return;
      }
      else
      {
        deError = getLastError();
        if( dynlen(deError) > 0 )
        {
          _unSystemIntegrity_IEC104_sendException("_unSystemIntegrity_IEC104_Checking", "ERROR", "Error getting the driver number, IEC connection and IEC driver " + sFrontEndDp + ".communication.driver_num" + ", due to: " + getErrorText(deError));
          return;
        }
      }

      sFrontEndDriver = "_Iec_" + iDriverNumber;
      if( dpExists(sFrontEndDriver) == FALSE )
      {
        _unSystemIntegrity_IEC104_sendException("_unSystemIntegrity_IEC104_Checking", "ERROR", "Error front end driver: " + sFrontEndDriver + " doesn't exist");
        return;
      }


      iThreadId = startThread(UN_SYSTEM_INTEGRITY_IEC104_check, sFrontEndName, iDriverNumber, sFrontEndConnection, sFrontEndDriver);
      g_diUnSystemIntegrity_IEC104_FrontEndThreadId[iPos] = iThreadId;
      if( iThreadId < 0)
      {
        iPos = _unSystemIntegrity_isInList(g_dsUnSystemIntegrity_IEC104_FrontEndList, sDp);
        if( iPos > 0 )
        {
          // Kill the threads if they were started
          iRes = stopThread(g_unSystemIntegrity_IEC104_DSThreadId[iPos]);
          if( iRes == -1 )
          {
            _unSystemIntegrity_IEC104_sendException("_unSystemIntegrity_IEC104_Checking", "ERROR", "Error stopping thread for FrontEnd: " + g_dsUnSystemIntegrity_IEC104_FrontEndList[iPos] + ", with ID: " + g_diUnSystemIntegrity_IEC104_FrontEndThreadId[iPos]);
            return;
          }

          // And then remove it from the list of FrontEnds and the list of FrontEnd Thread Id
          iPos = _unSystemIntegrity_setList(g_dsUnSystemIntegrity_IEC104_FrontEndList, sDp, FALSE);
          dynRemove(g_i_unSystemIntegrity_IEC104_CheckingDelay, iPos);
        }
        _unSystemIntegrity_IEC104_sendException("_unSystemIntegrity_IEC104_Checking", "ERROR", "Error " + getCatStr("unSystemIntegrity", "CANNOT_DP_CONNECT") + sDp, "");
        return;
      }
      else
      {
        // Set the enable to TRUE and activate the alarm if it is not activated.
        if( dpSet(sDp + ".enabled", TRUE) != 0 )
        {
          _unSystemIntegrity_IEC104_sendException("_unSystemIntegrity_IEC104_Checking", "ERROR", "Error enabling system integrity alarm: " + sDp);
          return;
        }
        else
        {
          deError = getLastError();
          if( dynlen(deError) > 0 )
          {
            _unSystemIntegrity_IEC104_sendException("_unSystemIntegrity_IEC104_Checking", "ERROR", "Error enabling system integrity alarm: " + sDp + ", due to: " + getErrorText(deError));
            return;
          }

          unAlarmConfig_mask(sDp + ".alarm", FALSE, exceptionInfo);
          if( dynlen(exceptionInfo) > 0 )
          {
            _unSystemIntegrity_IEC104_sendException("_unSystemIntegrity_IEC104_Checking", "ERROR", "Error unmasking system integrity alarm: " + sDp + ", due to: " + exceptionInfo[2]);
            return;
          }
        }
      }
    }
  }
  else
  {
    // Disconnect the callback function
    iPos = _unSystemIntegrity_isInList(g_dsUnSystemIntegrity_IEC104_FrontEndList, sDp);
    if( iPos > 0 )
    {
      // Kill the thread
      iRes = stopThread(g_diUnSystemIntegrity_IEC104_FrontEndThreadId[iPos]);
      if( iRes < 0 )
      {
        _unSystemIntegrity_IEC104_sendException("_unSystemIntegrity_IEC104_Checking", "ERROR", "Error stopping system integrity thread for: " + sDp + ", with Id: " + g_diUnSystemIntegrity_IEC104_FrontEndThreadId[iPos]);
        return;
      }


      // Remove from list
      iPos = _unSystemIntegrity_setList(g_dsUnSystemIntegrity_IEC104_FrontEndList, sDp, FALSE);
      dynRemove(g_diUnSystemIntegrity_IEC104_FrontEndThreadId, iPos);

      // Set the .enable to FALSE and de-activate the alarm and acknowledge it if necessary
      if( dpSet(sDp + ".enabled", FALSE,
                sDp + ".alarm",   c_unSystemIntegrity_no_alarm_value) != 0 )
      {
        _unSystemIntegrity_IEC104_sendException("_unSystemIntegrity_IEC104_Checking", "ERROR", "Error disabling and removing error condition from system integrity alarm: " + sDp);
        return;
      }
      else
      {
        deError = getLastError();
        if( dynlen(deError) > 0 )
        {
          _unSystemIntegrity_IEC104_sendException("_unSystemIntegrity_IEC104_Checking", "ERROR", "Error disabling and removing error condition from system integrity alarm: " + sDp + " due to: " + getErrorText(deError));
          return;
        }
      }

      unAlarmConfig_mask(sDp + ".alarm", true, exceptionInfo);
      if( dynlen(exceptionInfo) > 0 )
      {
        _unSystemIntegrity_IEC104_sendException("_unSystemIntegrity_IEC104_Checking", "ERROR", "Error masking alarm for system integrity: " + sDp + ", due to: " + exceptionInfo[2]);
        return;
      }
    }
  }
}//  _unSystemIntegrity_IEC104_Checking()





// _unSystemIntegrity_IEC104_Check
/**
Purpose: This function checks if the counter of the OPCUA dp is modified periodically. If not it generates and alarm.

@param sDp: string, input, data point name
@param iDriverNumber: int, input, driver number
@param sFrontEndConnection, string, input, Front end IEC104 connection
@param sFrontEndDriver: string, input, Front end IEC104 driver

Return: nothing
*/
void _unSystemIntegrity_IEC104_Check(string sDp, int iDriverNumber, string sFrontEndConnection, string sFrontEndDriver)
{
  bool bHostConnection, bDeviceConnectionsOk;
  int iWaitingTime, iOldAlarmValue, iAlarmValue, sDriverType, iConnState, iLoop, iLen;
  string sDriverType;
  dyn_int diNumber, diConnState;
  dyn_errClass deError;


  while( TRUE )
  {
    bDeviceConnectionsOk = TRUE;

    // If waiting time is 0, set it to g_i_unSystemIntegrity_IEC104_CheckingDelay
    iWaitingTime = g_i_unSystemIntegrity_IEC104_CheckingDelay;
    if( iWaitingTime <= 0 )
    {
      iWaitingTime = c_unSystemIntegrity_IEC104_defaultCheckingDelay;
    }
    delay (iWaitingTime);

    // Check functionality
    if( dpGet ("_Driver" + iDriverNumber + ".DT",             sDriverType,
               "_Connections.Driver.ManNums",                 diNumber,
               sFrontEndConnection + ".State",                bHostConnection,
               sFrontEndConnection + ".States.ConnState",     iConnState,
               sFrontEndConnection + ".States.AllConnStates", diConnState ) != 0 )
    {
      DebugTN("_unSystemIntegrity_IEC104_Check() -> ERROR -> Getting values from DPs to calculate connection status");
      continue;
    }
    else
    {
      deError = getLastError();
      if( dynlen(deError) > 0 )
      {
        DebugTN("_unSystemIntegrity_IEC104_Check() -> ERROR -> Getting values from DPs to calculate connection status, due to: " + getErrorText(deError));
        continue;
      }
    }


    if( sDriverType == "IEC" )
    {
      if( dynContains(diNumber, iDriverNumber) <= 0 )
      {
        iAlarmValue = iSYSTEM_INTEGRITY_IEC104_ERROR_MANAGER_STOPPED ;
      }
      else
      {
        if( bHostConnection == FALSE )
        {
          iAlarmValue = iSYSTEM_INTEGRITY_IEC104_ERROR_CONNECTION_TO_HOST ;
        }
        else
        {
          if( iConnState != 1 )
          {
            iAlarmValue = iSYSTEM_INTEGRITY_IEC104_ERROR_CONNECTION_NOT_READY ;
          }
          else
          {
            iLen = dynlen (diConnState);
            for( iLoop = 1 ; iLoop <= iLen ; iLoop++ )
            {
              if( diConnState[iLoop] != 1 )
              {
                iAlarmValue           = iSYSTEM_INTEGRITY_IEC104_ERROR_CONNECTION_DEVICE_NOT_READY ;
                bDeviceConnectionsOk  = FALSE;
                break; // break for loop
              }
            }

            if( bDeviceConnectionsOk == TRUE )
            {
              iAlarmValue = c_unSystemIntegrity_no_alarm_value;
            }
          }
        }
      }
    }
    else
    {
      iAlarmValue = iSYSTEM_INTEGRITY_IEC104_ERROR_DRIVER_TYPE;
    }


    if( iOldAlarmValue != iAlarmValue )
    {
      if( dpSet(c_unSystemAlarm_dpPattern + DS_pattern + c_unSystemIntegrity_IEC104 + sDp + ".alarm", iAlarmValue) != 0 )
      {
        _unSystemIntegrity_IEC104_sendException("_unSystemIntegrity_IEC104_Check", "ERROR", "Error setting the alarm: " + iAlarmValue + ", for system integrity: " + sDp);
        continue;
      }
      else
      {
        deError = getLastError();
        if( dynlen(deError) > 0 )
        {
          _unSystemIntegrity_IEC104_sendException("_unSystemIntegrity_IEC104_Check", "ERROR", "Error setting the alarm: " + iAlarmValue + ", for system integrity: " + sDp + " due to: " + getErrorText(deError));
          continue;
        }
      }
      iOldAlarmValue = iAlarmValue;
    }

  }

}//  _unSystemIntegrity_IEC104_Check()





